# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import _cython_3_0_11
import nanoarrow._device
from _typeshed import Incomplete
from typing import ClassVar

DEVICE_CPU: nanoarrow._device.Device
__reduce_cython__: _cython_3_0_11.cython_function_or_method
__setstate_cython__: _cython_3_0_11.cython_function_or_method
__test__: dict

class CBuffer:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    data_type: Incomplete
    data_type_id: Incomplete
    device: Incomplete
    element_size_bits: Incomplete
    format: Incomplete
    itemsize: Incomplete
    n_elements: Incomplete
    size_bytes: Incomplete
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    def element(self, *args, **kwargs): ...
    def elements(self, *args, **kwargs): ...
    @staticmethod
    def empty(*args, **kwargs):
        """Create an empty CBuffer"""
    @staticmethod
    def from_dlpack(*args, **kwargs):
        """Create a CBuffer using the DLPack protocol

        Wraps a tensor from an external library as a CBuffer that can be used
        to create an array.

        Parameters
        ----------
        obj : object with a ``__dlpack__`` attribute
            The object on which to invoke the DLPack protocol
        stream : int, optional
            The stream on which the tensor represented by obj should be made
            safe for use. This value is passed to the object's ``__dlpack__``
            method; however, the CBuffer does not keep any record of this (i.e.,
            the caller is responsible for creating a sync event after creating one
            or more buffers in this way).
        """
    @staticmethod
    def from_pybuffer(*args, **kwargs):
        """Create a CBuffer using the Python buffer protocol

        Wraps a buffer using the Python buffer protocol as a CBuffer that can be
        used to create an array.

        Parameters
        ----------
        obj : buffer-like
            The object on which to invoke the Python buffer protocol
        """
    def view(self, *args, **kwargs):
        """Export this buffer as a CBufferView

        Returns a :class:`CBufferView` of this buffer. After calling this
        method, the original CBuffer will be invalidated and cannot be used.
        In general, the view of the buffer should be used to consume a buffer
        (whereas the CBuffer is primarily used to wrap an existing object in
        a way that it can be used to build a :class:`CArray`).
        """
    def __buffer__(self, *args, **kwargs):
        """Return a buffer object that exposes the underlying memory of the object."""
    def __getitem__(self, index):
        """Return self[key]."""
    def __iter__(self):
        """Implement iter(self)."""
    def __len__(self) -> int:
        """Return len(self)."""
    def __reduce__(self): ...
    def __release_buffer__(self, *args, **kwargs):
        """Release the buffer object that exposes the underlying memory of the object."""

class CBufferBuilder:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    capacity_bytes: Incomplete
    format: Incomplete
    itemsize: Incomplete
    size_bytes: Incomplete
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    def advance(self, *args, **kwargs):
        """Manually increase :attr:`size_bytes` by ``additional_bytes``

        This can be used after writing to the buffer using the buffer protocol
        to ensure that :attr:`size_bytes` accurately reflects the number of
        bytes written to the buffer.
        """
    def finish(self, *args, **kwargs):
        """Finish building this buffer

        Performs any steps required to finish building this buffer and
        returns the result. Any behaviour resulting from calling methods
        on this object after it has been finished is not currently
        defined (but should not crash).
        """
    def reserve_bytes(self, *args, **kwargs):
        """Ensure that the underlying buffer has space for ``additional_bytes``
        more bytes to be written"""
    def set_data_type(self, *args, **kwargs):
        """Set the data type used to interpret elements in :meth:`write_elements`."""
    def set_format(self, *args, **kwargs):
        """Set the Python buffer format used to interpret elements in
        :meth:`write_elements`.
        """
    def write(self, *args, **kwargs):
        """Write bytes to this buffer

        Writes the bytes of ``content`` without considering the element type of
        ``content`` or the element type of this buffer.

        This method returns the number of bytes that were written.
        """
    def write_elements(self, *args, **kwargs):
        """ "Write an iterable of elements to this buffer

        Writes the elements of iterable ``obj`` according to the binary
        representation specified by :attr:`format`. This is currently
        powered by ``struct.pack_into()`` except when building bitmaps
        where an internal implementation is used.

        This method returns the number of elements that were written.
        """
    def write_fill(self, *args, **kwargs):
        """Write fill bytes to this buffer

        Appends the byte ``value`` to this buffer ``size_bytes`` times.
        """
    def __buffer__(self, *args, **kwargs):
        """Return a buffer object that exposes the underlying memory of the object."""
    def __len__(self) -> int:
        """Return len(self)."""
    def __reduce__(self): ...
    def __release_buffer__(self, *args, **kwargs):
        """Release the buffer object that exposes the underlying memory of the object."""

class CBufferView:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    data_type: Incomplete
    data_type_id: Incomplete
    device: Incomplete
    element_size_bits: Incomplete
    format: Incomplete
    itemsize: Incomplete
    n_elements: Incomplete
    size_bytes: Incomplete
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    def copy(self, *args, **kwargs): ...
    def copy_into(self, *args, **kwargs): ...
    def element(self, *args, **kwargs): ...
    def elements(self, *args, **kwargs): ...
    def unpack_bits(self, *args, **kwargs): ...
    def unpack_bits_into(self, *args, **kwargs): ...
    def __buffer__(self, *args, **kwargs):
        """Return a buffer object that exposes the underlying memory of the object."""
    def __dlpack__(self, *args, **kwargs):
        """
        Export CBufferView as a DLPack capsule.

        Parameters
        ----------
        stream : int, optional
            A Python integer representing a pointer to a stream.
            Stream is provided by the consumer to the producer to instruct the producer
            to ensure that operations can safely be performed on the array.

        Returns
        -------
        capsule : PyCapsule
            A DLPack capsule for the array, pointing to a DLManagedTensor.
        """
    def __dlpack_device__(self, *args, **kwargs):
        """
        Return the DLPack device tuple this CBufferView resides on.

        Returns
        -------
        tuple : Tuple[int, int]
            Tuple with index specifying the type of the device (where
            CPU = 1, see python/src/nanoarrow/dpack_abi.h) and index of the
            device which is 0 by default for CPU.
        """
    def __getitem__(self, index):
        """Return self[key]."""
    def __iter__(self):
        """Implement iter(self)."""
    def __len__(self) -> int:
        """Return len(self)."""
    def __reduce__(self): ...
    def __release_buffer__(self, *args, **kwargs):
        """Release the buffer object that exposes the underlying memory of the object."""

class NoneAwareWrapperIterator:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    def finish(self, *args, **kwargs):
        """Obtain the total count, null count, and validity bitmap after
        consuming this iterable."""
    def reserve(self, *args, **kwargs): ...
    def __iter__(self):
        """Implement iter(self)."""
    def __reduce__(self): ...



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from nanoarrow._buffer import CBuffer, CBufferBuilder
from nanoarrow._utils import obj_is_buffer
from nanoarrow.c_schema import c_schema_view

from nanoarrow import _types


def c_buffer(obj, schema=None) -> CBuffer:
    """Owning, read-only ArrowBuffer wrapper

    If obj implements the Python buffer protocol, ``c_buffer()`` wraps
    obj in nanoarrow's owning buffer structure, the ArrowBuffer,
    such that it can be used to construct arrays. The ownership of the
    underlying buffer is handled by the Python buffer protocol
    (i.e., ``PyObject_GetBuffer()`` and ``PyBuffer_Release()``).

    If obj is iterable, a buffer will be allocated and populated with
    the contents of obj according to ``schema``. The
    ``schema`` parameter is required to create a buffer from
    a Python iterable. The ``struct`` module is currently used to encode
    values from obj into binary form.

    Unlike with :func:`c_array`, ``schema`` is explicitly
    honoured (or an error will be raised).

    Parameters
    ----------

    obj : buffer-like, tensor, or iterable
        A Python object that supports the Python buffer or DLPack protocols.
        This includes bytes, memoryview, bytearray, bulit-in types as well
        as numpy arrays.
    schema :  schema-like, optional
        The data type of the desired buffer as sanitized by
        :func:`c_schema`. Only values that make sense as buffer types are
        allowed (e.g., integer types, floating-point types, interval types,
        decimal types, binary, string, fixed-size binary).

    Examples
    --------

    >>> import nanoarrow as na
    >>> na.c_buffer(b"1234")
    nanoarrow.c_buffer.CBuffer(uint8[4 b] 49 50 51 52)
    >>> na.c_buffer([1, 2, 3], na.int32())
    nanoarrow.c_buffer.CBuffer(int32[12 b] 1 2 3)
    """
    if isinstance(obj, CBuffer) and schema is None:
        return obj

    if obj_is_buffer(obj):
        if schema is not None:
            raise NotImplementedError(
                "c_buffer() with schema for pybuffer is not implemented"
            )
        return CBuffer.from_pybuffer(obj)

    if _obj_is_tensor(obj):
        if schema is not None:
            raise NotImplementedError(
                "c_buffer() with schema for DLPack is not implemented"
            )
        return CBuffer.from_dlpack(obj)

    if _obj_is_iterable(obj):
        buffer, _ = _c_buffer_from_iterable(obj, schema)
        return buffer

    raise TypeError(
        f"Can't convert object of type {type(obj).__name__} to nanoarrow.c_buffer"
    )


def _c_buffer_from_iterable(obj, schema=None) -> CBuffer:
    import array

    # array.typecodes is not available in all PyPy versions.
    # Rather than guess, just don't use the array constructor if
    # this attribute is not available.
    if hasattr(array, "typecodes"):
        array_typecodes = array.typecodes
    else:
        array_typecodes = []

    if schema is None:
        raise ValueError("CBuffer from iterable requires schema")

    schema_view = c_schema_view(schema)
    if (
        schema_view.extension_name is not None
        or schema_view.storage_type_id != schema_view.type_id
    ):
        raise ValueError(
            f"Can't create buffer from iterable for type {schema_view.type}"
        )

    builder = CBufferBuilder()

    if schema_view.storage_type_id == _types.FIXED_SIZE_BINARY:
        builder.set_data_type(_types.BINARY, schema_view.fixed_size * 8)
    else:
        builder.set_data_type(schema_view.storage_type_id)

    # If we are using a typecode supported by the array module, it has much
    # faster implementations of safely building buffers from iterables
    if builder.format in array_typecodes and schema_view.storage_type_id != _types.BOOL:
        buf = array.array(builder.format, obj)
        return CBuffer.from_pybuffer(buf), len(buf)

    n_values = builder.write_elements(obj)
    return builder.finish(), n_values


def _obj_is_iterable(obj):
    return hasattr(obj, "__iter__")


def _obj_is_tensor(obj):
    return hasattr(obj, "__dlpack__")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import _cython_3_0_11
from _typeshed import Incomplete

__reduce_cython__: _cython_3_0_11.cython_function_or_method
__setstate_cython__: _cython_3_0_11.cython_function_or_method
__test__: dict
init_array_stream: _cython_3_0_11.cython_function_or_method

class CIpcInputStream:
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    @staticmethod
    def from_readable(*args, **kwargs): ...
    def is_valid(self, *args, **kwargs): ...
    def release(self, *args, **kwargs): ...
    def __reduce__(self): ...

class CIpcOutputStream:
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    @staticmethod
    def from_writable(*args, **kwargs): ...
    def is_valid(self, *args, **kwargs): ...
    def release(self, *args, **kwargs): ...
    def __reduce__(self): ...

class CIpcWriter:
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    def is_valid(self, *args, **kwargs): ...
    def release(self, *args, **kwargs): ...
    def write_array_stream(self, *args, **kwargs): ...
    def write_array_view(self, *args, **kwargs): ...
    def write_end_of_stream(self, *args, **kwargs): ...
    def write_schema(self, *args, **kwargs): ...
    def __reduce__(self): ...

class PyStreamPrivate:
    close_obj: Incomplete
    obj: Incomplete
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    def set_buffer(self, *args, **kwargs): ...
    def __buffer__(self, *args, **kwargs):
        """Return a buffer object that exposes the underlying memory of the object."""
    def __len__(self) -> int:
        """Return len(self)."""
    def __reduce__(self): ...
    def __release_buffer__(self, *args, **kwargs):
        """Release the buffer object that exposes the underlying memory of the object."""



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import pytest
from nanoarrow.c_schema import allocate_c_schema, c_schema_view

import nanoarrow as na


def test_c_schema_basic():
    schema = allocate_c_schema()
    assert schema.is_valid() is False
    assert schema._to_string() == "[invalid: schema is released]"
    assert repr(schema) == "<nanoarrow.c_schema.CSchema <released>>"

    schema = na.c_schema(na.struct({"some_name": na.int32()}))

    assert schema.format == "+s"
    assert schema.flags == 2
    assert schema.metadata is None
    assert schema.n_children == 1
    assert len(list(schema.children)) == 1
    assert schema.child(0).format == "i"
    assert schema.child(0).name == "some_name"
    assert schema.child(0)._to_string() == "int32"
    assert "<nanoarrow.c_schema.CSchema int32>" in repr(schema)
    assert schema.dictionary is None

    with pytest.raises(IndexError):
        schema.child(1)


def test_c_schema_dictionary():
    pa = pytest.importorskip("pyarrow")

    schema = na.c_schema(pa.dictionary(pa.int32(), pa.utf8()))
    assert schema.format == "i"
    assert schema.dictionary.format == "u"
    assert "dictionary: <nanoarrow.c_schema.CSchema string" in repr(schema)


def test_schema_metadata():
    meta = {"key1": "value1", "key2": "value2"}
    schema = na.c_schema(na.int32()).modify(metadata=meta)

    assert len(schema.metadata) == 2

    meta2 = {k: v for k, v in schema.metadata.items()}
    assert list(meta2.keys()) == [b"key1", b"key2"]
    assert list(meta2.values()) == [b"value1", b"value2"]
    assert "b'key1': b'value1'" in repr(schema)


def test_c_schema_view():
    schema = allocate_c_schema()
    with pytest.raises(RuntimeError):
        c_schema_view(schema)

    schema = na.c_schema(na.int32())
    view = c_schema_view(schema)
    assert "- type: 'int32'" in repr(view)
    assert view.type == "int32"
    assert view.storage_type == "int32"

    assert view.fixed_size is None
    assert view.decimal_bitwidth is None
    assert view.decimal_scale is None
    assert view.time_unit is None
    assert view.timezone is None
    assert view.union_type_ids is None
    assert view.extension_name is None
    assert view.extension_metadata is None


def test_c_schema_view_extra_params():
    view = c_schema_view(na.fixed_size_binary(12))
    assert view.fixed_size == 12

    view = c_schema_view(na.decimal128(10, 3))
    assert view.decimal_bitwidth == 128
    assert view.decimal_precision == 10
    assert view.decimal_scale == 3

    view = c_schema_view(na.decimal256(10, 3))
    assert view.decimal_bitwidth == 256
    assert view.decimal_precision == 10
    assert view.decimal_scale == 3

    view = c_schema_view(na.duration("us"))
    assert view.time_unit == "us"

    view = c_schema_view(na.timestamp("us", "America/Halifax"))
    assert view.type == "timestamp"
    assert view.storage_type == "int64"
    assert view.time_unit == "us"
    assert view.timezone == "America/Halifax"

    pa = pytest.importorskip("pyarrow")

    view = c_schema_view(pa.list_(pa.int32(), 12))
    assert view.fixed_size == 12


def test_c_schema_metadata():
    meta = {
        b"ARROW:extension:name": b"some_name",
        b"ARROW:extension:metadata": b"some_metadata",
    }

    schema = na.c_schema(na.int32()).modify(metadata=meta)
    assert "b'some_name'" in repr(schema)
    assert "b'some_name'" in repr(schema.metadata)
    assert list(schema.metadata) == list(meta)
    assert list(schema.metadata.items()) == list(meta.items())
    assert list(schema.metadata.keys()) == list(meta.keys())
    assert list(schema.metadata.values()) == list(meta.values())

    view = c_schema_view(schema)
    assert view.extension_name == "some_name"
    assert view.extension_metadata == b"some_metadata"


def test_c_schema_equals():
    int32 = na.c_schema(na.int32())
    struct = na.c_schema(na.struct({"col1": na.int32()}))
    dictionary = na.c_schema(na.dictionary(na.int32(), na.string()))
    ordered_dictionary = na.c_schema(
        na.dictionary(na.int32(), na.string(), dictionary_ordered=True)
    )

    # Check schemas pointing to the same ArrowSchema
    assert int32.type_equals(int32)

    # Check equality with deep copies
    assert int32.type_equals(int32.__deepcopy__())
    assert struct.type_equals(struct.__deepcopy__())
    assert dictionary.type_equals(dictionary.__deepcopy__())

    # Check inequality because of format
    assert int32.type_equals(struct) is False

    # Check inequality because of nullability
    assert int32.type_equals(int32.modify(flags=0), check_nullability=True) is False
    # ...but not by default
    assert int32.type_equals(int32.modify(flags=0)) is True

    # Check inequality of type information encoded in flags
    assert dictionary.type_equals(ordered_dictionary) is False

    # Check inequality because of number of children
    assert struct.type_equals(struct.modify(children=[])) is False

    # Check inequality because of a difference in the children
    assert struct.type_equals(struct.modify(children=[dictionary])) is False

    # Check inequality because of dictionary presence
    assert int32.type_equals(dictionary) is False
    assert dictionary.type_equals(int32) is False

    # Check inequality because of dictionary index type
    assert (
        dictionary.type_equals(na.c_schema(na.dictionary(na.int64(), na.string())))
        is False
    )

    # Check inequality because of dictionary value type
    assert dictionary.type_equals(dictionary.modify(dictionary=struct)) is False


def test_c_schema_assert_type_equal():
    from nanoarrow._schema import assert_type_equal

    int32 = na.c_schema(na.int32())
    string = na.c_schema(na.string())
    nn_string = na.c_schema(na.string(False))

    assert_type_equal(int32, int32, check_nullability=True)

    with pytest.raises(TypeError):
        assert_type_equal(None, int32, check_nullability=False)

    with pytest.raises(TypeError):
        assert_type_equal(int32, None, check_nullability=False)

    msg = "Expected schema\n  'string'\nbut got\n  'int32'"
    with pytest.raises(ValueError, match=msg):
        assert_type_equal(int32, string, check_nullability=False)

    assert_type_equal(nn_string, string, check_nullability=False)
    with pytest.raises(ValueError):
        assert_type_equal(nn_string, string, check_nullability=True)


def test_c_schema_modify():
    schema = na.c_schema(na.null())

    schema_clone = schema.modify()
    assert schema_clone is not schema
    assert schema._addr() != schema_clone._addr()

    schema_formatted = schema.modify(format="i")
    assert schema_formatted.format == "i"

    schema_named = schema.modify(name="something else")
    assert schema_named.name == "something else"
    assert schema_named.format == schema.format

    schema_flagged = schema.modify(flags=0)
    assert schema_flagged.flags == 0
    assert schema_flagged.format == schema.format

    schema_non_nullable = schema.modify(nullable=False)
    assert schema_non_nullable.flags == 0
    assert schema_non_nullable.format == schema.format

    meta = {"some key": "some value"}
    schema_metad = schema.modify(metadata=meta)
    assert list(schema_metad.metadata.items()) == [(b"some key", b"some value")]
    assert schema_non_nullable.format == schema.format

    schema_metad2 = schema.modify(metadata=schema_metad.metadata)
    assert list(schema_metad2.metadata.items()) == [(b"some key", b"some value")]

    schema_no_metad = schema_metad.modify(metadata={})
    assert schema_no_metad.metadata is None


def test_c_schema_modify_children():
    schema = na.c_schema(na.struct({"col1": na.null()}))

    schema_same_children = schema.modify()
    assert schema_same_children.n_children == 1
    assert schema_same_children.child(0).name == "col1"
    assert schema_same_children.child(0).format == "n"

    schema_new_children_list = schema.modify(
        children=[na.c_schema(na.int32()).modify(name="new name")]
    )
    assert schema_new_children_list.n_children == 1
    assert schema_new_children_list.child(0).name == "new name"
    assert schema_new_children_list.child(0).format == "i"

    schema_new_children_dict = schema.modify(
        children={"new name": na.c_schema(na.int32())}
    )
    assert schema_new_children_dict.n_children == 1
    assert schema_new_children_dict.child(0).name == "new name"
    assert schema_new_children_dict.child(0).format == "i"


def test_c_schema_modify_dictionary():
    schema = na.c_schema(na.int32())

    schema_dictionary = schema.modify(dictionary=na.c_schema(na.string()))
    assert schema_dictionary.format == "i"
    assert schema_dictionary.dictionary.format == "u"

    schema_same_dictionary = schema_dictionary.modify()
    assert schema_same_dictionary.format == "i"
    assert schema_same_dictionary.dictionary.format == "u"

    schema_no_dictionary = schema_dictionary.modify(dictionary=False)
    assert schema_no_dictionary.format == "i"
    assert schema.dictionary is None



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
import pathlib
import shutil


def main():
    src_dir = pathlib.Path(os.environ["MESON_SOURCE_ROOT"]).parent.resolve()
    dist_dir = pathlib.Path(os.environ["MESON_DIST_ROOT"]).resolve()
    subproj_dir = dist_dir / "subprojects" / "arrow-nanoarrow"

    if subproj_dir.is_symlink():
        subproj_dir.unlink()

    subproj_dir.mkdir(parents=True)
    shutil.copy(src_dir / "meson.build", subproj_dir / "meson.build")
    shutil.copy(src_dir / "meson_options.txt", subproj_dir / "meson_options.txt")

    # Copy over any subproject dependency / wrap files
    subproj_subproj_dir = subproj_dir / "subprojects"
    subproj_subproj_dir.mkdir()
    for f in (src_dir / "subprojects").glob("*.wrap"):
        shutil.copy(f, subproj_subproj_dir / f.name)
    shutil.copytree(
        src_dir / "subprojects" / "packagefiles", subproj_subproj_dir / "packagefiles"
    )

    target_src_dir = subproj_dir / "src"
    shutil.copytree(src_dir / "src", target_src_dir)

    # CMake isn't actually required for building, but the bundle.py script reads from
    # its configuration
    shutil.copy(src_dir / "CMakeLists.txt", subproj_dir / "CMakeLists.txt")

    subproj_ci_scripts_dir = subproj_dir / "ci" / "scripts"
    subproj_ci_scripts_dir.mkdir(parents=True)
    shutil.copy(
        src_dir / "ci" / "scripts" / "bundle.py", subproj_ci_scripts_dir / "bundle.py"
    )


if __name__ == "__main__":
    main()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# cython: language_level = 3

from libc.stdint cimport uint8_t, int64_t

from nanoarrow_c cimport (
    ArrowSchema,
    ArrowArray,
    ArrowArrayStream,
    ArrowArrayView,
    ArrowBuffer,
    ArrowError
)
from nanoarrow_device_c cimport ArrowDeviceArray

cdef object alloc_c_schema(ArrowSchema** c_schema)

cdef object alloc_c_array(ArrowArray** c_array)

cdef object alloc_c_array_stream(ArrowArrayStream** c_stream)

cdef object alloc_c_device_array(ArrowDeviceArray** c_device_array)

cdef object alloc_c_array_view(ArrowArrayView** c_array_view)

cdef object alloc_c_buffer(ArrowBuffer** c_buffer)

cdef void c_array_shallow_copy(object base, const ArrowArray* src, ArrowArray* dst)

cdef void c_device_array_shallow_copy(object base, const ArrowDeviceArray* src,
                                      ArrowDeviceArray* dst)

cdef object c_buffer_set_pybuffer(object obj, ArrowBuffer** c_buffer)

cdef void c_buffer_set_pyobject(object base, uint8_t* data, int64_t size_bytes, ArrowBuffer** c_buffer)

cdef class Error:
    cdef ArrowError c_error

    cdef raise_message(self, what, code)

    cdef raise_message_not_ok(self, what, code)

    @staticmethod
    cdef raise_error(what, code)

    @staticmethod
    cdef raise_error_not_ok(what, code)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import re
import sys

import pytest
from nanoarrow.c_array import c_array_view

import nanoarrow as na

np = pytest.importorskip("numpy")
pa = pytest.importorskip("pyarrow")


def test_c_version():
    re_version = re.compile(r"^[0-9]+\.[0-9]+\.[0-9]+(-SNAPSHOT)?$")
    assert re_version.match(na.c_version()) is not None


def test_c_schema_helper():
    from nanoarrow.c_schema import CSchema, allocate_c_schema

    schema = allocate_c_schema()
    assert na.c_schema(schema) is schema

    schema = na.c_schema(pa.null())
    assert isinstance(schema, CSchema)

    with pytest.raises(TypeError):
        na.c_schema(1234)


def test_c_array_helper():
    from nanoarrow.c_array import CArray, allocate_c_array

    array = allocate_c_array()
    assert na.c_array(array) is array

    array = na.c_array(pa.array([], pa.null()))
    assert isinstance(array, CArray)

    with pytest.raises(TypeError):
        na.c_array(1234)


def test_array_stream_helper():
    from nanoarrow.c_array_stream import allocate_c_array_stream

    array_stream = allocate_c_array_stream()
    assert na.c_array_stream(array_stream) is array_stream

    with pytest.raises(TypeError):
        na.c_array_stream(1234)


def test_array_view_helper():
    from nanoarrow.c_array import CArrayView, c_array_view

    array = na.c_array(pa.array([1, 2, 3]))
    view = c_array_view(array)
    assert isinstance(view, CArrayView)
    assert c_array_view(view) is view


def test_c_array_empty():
    from nanoarrow.c_array import allocate_c_array

    array = allocate_c_array()
    assert array.is_valid() is False
    assert repr(array) == "<nanoarrow.c_array.CArray <released>>"


def test_c_array():
    array = na.c_array(pa.array([1, 2, 3], pa.int32()))
    assert array.is_valid() is True
    assert array.length == 3
    assert len(array) == 3
    assert array.offset == 0
    assert array.null_count == 0
    assert array.n_buffers == 2
    assert len(array.buffers) == 2
    assert array.buffers[0] == 0
    assert array.n_children == 0
    assert len(list(array.children)) == 0
    assert array.dictionary is None
    assert "<nanoarrow.c_array.CArray int32" in repr(array)


def test_c_array_recursive():
    array = na.c_array(pa.record_batch([pa.array([1, 2, 3], pa.int32())], ["col"]))
    assert array.n_children == 1
    assert len(list(array.children)) == 1
    assert array.child(0).length == 3
    assert array.child(0).schema._to_string() == "int32"
    assert "'col': <nanoarrow.c_array.CArray int32" in repr(array)

    with pytest.raises(IndexError):
        array.child(-1)


def test_c_array_dictionary():
    array = na.c_array(pa.array(["a", "b", "b"]).dictionary_encode())
    assert array.length == 3
    assert array.dictionary.length == 2
    assert "dictionary: <nanoarrow.c_array.CArray string>" in repr(array)


def test_c_array_view():
    array = na.c_array(pa.array([1, 2, 3], pa.int32()))
    view = array.view()

    assert view.storage_type == "int32"
    assert "- storage_type: 'int32'" in repr(view)
    assert "data <int32[12 b] 1 2 3>" in repr(view)

    data_buffer = memoryview(view.buffer(1))
    data_buffer_copy = bytes(data_buffer)
    assert len(data_buffer_copy) == 12

    if sys.byteorder == "little":
        assert data_buffer_copy == b"\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00"
    else:
        assert data_buffer_copy == b"\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x00\x03"

    with pytest.raises(IndexError):
        view.child(0)

    with pytest.raises(IndexError):
        view.child(-1)


def test_c_array_view_recursive():
    pa_array_child = pa.array([1, 2, 3], pa.int32())
    pa_array = pa.record_batch([pa_array_child], names=["some_column"])

    array = na.c_array(pa_array)

    assert array.schema.format == "+s"
    assert array.length == 3
    assert array.n_children == 1
    assert len(list(array.children)) == 1

    assert array.child(0).schema.format == "i"
    assert array.child(0).length == 3
    assert array.child(0).schema._addr() == array.schema.child(0)._addr()

    view = array.view()
    assert view.n_buffers == 1
    assert len(list(view.buffers)) == 1
    assert view.n_children == 1
    assert len(list(view.children)) == 1

    assert view.child(0).n_buffers == 2
    assert len(list(view.child(0).buffers)) == 2
    assert "- children[1]" in repr(view)


def test_c_array_view_dictionary():
    pa_array = pa.array(["a", "b", "b"], pa.dictionary(pa.int32(), pa.utf8()))
    array = na.c_array(pa_array)

    assert array.schema.format == "i"
    assert array.dictionary.schema.format == "u"

    view = array.view()
    assert view.n_buffers == 2
    assert view.dictionary.n_buffers == 3
    assert "- dictionary: <nanoarrow.c_array.CArrayView>" in repr(view)


def test_c_array_view_null_count():
    # With explicit null count == 0
    array = na.c_array_from_buffers(
        na.int32(), 3, (None, na.c_buffer([1, 2, 3], na.int32())), null_count=0
    )
    assert array.view().null_count == 0

    # Infer null count == 0 because of null data buffer when the null count
    # has not yet been computed by the producer.
    array = na.c_array_from_buffers(
        na.int32(), 3, (None, na.c_buffer([1, 2, 3], na.int32())), null_count=-1
    )
    assert array.view().null_count == 0

    # Compute null count == 0 by counting validity bits when the null count
    # has not yet been computed by the producer.
    array = na.c_array_from_buffers(
        na.int32(),
        3,
        (
            na.c_buffer([True, True, True], na.bool_()),
            na.c_buffer([1, 2, 3], na.int32()),
        ),
        null_count=-1,
    )

    assert array.view().null_count == 0

    # Check computed null count with actual nulls when the null count
    # has not yet been computed by the producer.
    array = na.c_array_from_buffers(
        na.int32(),
        3,
        (
            na.c_buffer([True, False, True], na.bool_()),
            na.c_buffer([1, 2, 3], na.int32()),
        ),
        null_count=-1,
    )
    assert array.view().null_count == 1


def test_buffers_integer():
    data_types = [
        (pa.uint8(), np.uint8()),
        (pa.int8(), np.int8()),
        (pa.uint16(), np.uint16()),
        (pa.int16(), np.int16()),
        (pa.uint32(), np.uint32()),
        (pa.int32(), np.int32()),
        (pa.uint64(), np.uint64()),
        (pa.int64(), np.int64()),
    ]

    for pa_type, np_type in data_types:
        view = c_array_view(pa.array([0, 1, 2], pa_type))
        data_buffer = view.buffer(1)

        # Check via buffer interface
        np.testing.assert_array_equal(
            np.array(data_buffer), np.array([0, 1, 2], np_type)
        )

        # Check via iterator interface
        assert list(data_buffer) == [0, 1, 2]

        # Check via buffer get_item interface
        assert [data_buffer[i] for i in range(len(data_buffer))] == list(data_buffer)

        # Check repr
        assert "0 1 2" in repr(data_buffer)


def test_buffers_float():
    data_types = [
        (pa.float32(), np.float32()),
        (pa.float64(), np.float64()),
    ]

    for pa_type, np_type in data_types:
        view = c_array_view(pa.array([0, 1, 2], pa_type))
        data_buffer = view.buffer(1)

        # Check via buffer interface
        np.testing.assert_array_equal(
            np.array(data_buffer), np.array([0, 1, 2], np_type)
        )

        # Check via iterator interface
        assert list(data_buffer) == [0.0, 1.0, 2.0]

        # Check via buffer get_item interface
        assert [data_buffer[i] for i in range(len(data_buffer))] == list(data_buffer)

        # Check repr
        assert "0.0 1.0 2.0" in repr(data_buffer)


def test_buffers_half_float():
    # pyarrrow can only create half_float from np.float16()
    np_array = np.array([0, 1, 2], np.float16())
    view = c_array_view(pa.array(np_array))
    data_buffer = view.buffer(1)

    # Check via buffer interface
    np.testing.assert_array_equal(
        np.array(data_buffer), np.array([0, 1, 2], np.float16())
    )

    # Check via iterator interface
    assert list(data_buffer) == [0.0, 1.0, 2.0]

    # Check via buffer get_item interface
    assert [data_buffer[i] for i in range(len(data_buffer))] == list(data_buffer)

    # Check repr
    assert "0.0 1.0 2.0" in repr(data_buffer)


def test_buffers_bool():
    view = c_array_view(pa.array([True, True, True, False]))
    data_buffer = view.buffer(1)

    assert data_buffer.size_bytes == 1

    # Check via buffer interface
    np.testing.assert_array_equal(
        np.array(data_buffer), np.array([1 + 2 + 4], np.int32())
    )

    # Check via iterator interface
    assert list(data_buffer) == [1 + 2 + 4]

    # Check via buffer get_item interface
    assert [data_buffer[i] for i in range(len(data_buffer))] == list(data_buffer)

    # Check via element interface
    assert data_buffer.n_elements == 8
    assert list(data_buffer.elements()) == [True] * 3 + [False] * 5
    assert [data_buffer.element(i) for i in range(data_buffer.n_elements)] == list(
        data_buffer.elements()
    )

    with pytest.raises(IndexError):
        data_buffer[8]
    with pytest.raises(IndexError):
        data_buffer[-1]
    with pytest.raises(IndexError):
        next(data_buffer.elements(-1, 4))
    with pytest.raises(IndexError):
        next(data_buffer.elements(7, 2))

    # Check repr
    assert "11100000" in repr(data_buffer)


def test_buffers_string():
    view = c_array_view(pa.array(["a", "bc", "def"]))

    assert view.buffer(0).size_bytes == 0
    assert view.buffer(1).size_bytes == 16
    assert view.buffer(2).size_bytes == 6

    # Check via buffer interface
    np.testing.assert_array_equal(
        np.array(view.buffer(1)), np.array([0, 1, 3, 6], np.int32())
    )
    np.testing.assert_array_equal(
        np.array(view.buffer(2)), np.array(list("abcdef"), dtype="|S1")
    )

    # Check via iterator interface
    assert list(view.buffer(0)) == []
    assert list(view.buffer(1)) == [0, 1, 3, 6]
    assert list(view.buffer(2)) == [item.encode("UTF-8") for item in "abcdef"]

    # Check repr
    assert "b'abcdef'" in repr(view.buffer(2))


def test_buffers_binary():
    view = c_array_view(pa.array([b"a", b"bc", b"def"]))

    assert view.buffer(0).size_bytes == 0
    assert view.buffer(1).size_bytes == 16
    assert view.buffer(2).size_bytes == 6

    # Check via buffer interface
    np.testing.assert_array_equal(
        np.array(view.buffer(1)), np.array([0, 1, 3, 6], np.int32())
    )
    np.testing.assert_array_equal(np.array(view.buffer(2)), np.array(list(b"abcdef")))
    np.testing.assert_array_equal(
        np.array(list(view.buffer(2))), np.array(list(b"abcdef"))
    )

    # Check via iterator interface
    assert list(view.buffer(0)) == []
    assert list(view.buffer(1)) == [0, 1, 3, 6]
    assert list(view.buffer(2)) == [int(item) for item in b"abcdef"]

    # Check repr
    assert "b'abcdef'" in repr(view.buffer(2))


def test_buffers_fixed_size_binary():
    view = c_array_view(pa.array([b"abc", b"def", b"ghi"], pa.binary(3)))

    assert view.buffer(1).size_bytes == 9

    # Check via buffer interface
    np.testing.assert_array_equal(
        np.array(list(view.buffer(1))), np.array([b"abc", b"def", b"ghi"])
    )

    # Check via iterator interface
    assert list(view.buffer(1)) == [b"abc", b"def", b"ghi"]


def test_buffers_interval_month_day_nano():
    view = c_array_view(
        pa.array([pa.scalar((1, 15, -30), type=pa.month_day_nano_interval())])
    )

    assert view.buffer(1).size_bytes == 16

    # Check via buffer interface
    np.testing.assert_array_equal(
        np.array(list(view.buffer(1))), np.array([(1, 15, -30)])
    )

    # Check via iterator interface
    assert list(view.buffer(1)) == [(1, 15, -30)]


def test_c_array_stream():
    from nanoarrow.c_array_stream import allocate_c_array_stream

    array_stream = allocate_c_array_stream()
    assert na.c_array_stream(array_stream) is array_stream
    assert repr(array_stream) == "<nanoarrow.c_array_stream.CArrayStream <released>>"

    assert array_stream.is_valid() is False
    with pytest.raises(RuntimeError):
        array_stream.get_schema()
    with pytest.raises(RuntimeError):
        array_stream.get_next()

    pa_array_child = pa.array([1, 2, 3], pa.int32())
    pa_array = pa.record_batch([pa_array_child], names=["some_column"])
    reader = pa.RecordBatchReader.from_batches(pa_array.schema, [pa_array])
    array_stream = na.c_array_stream(reader)

    assert array_stream.is_valid() is True
    assert "struct<some_column: int32>" in repr(array_stream)

    array = array_stream.get_next()
    assert array.schema.child(0).name == "some_column"
    with pytest.raises(StopIteration):
        array_stream.get_next()


def test_c_array_stream_iter():
    pa_array_child = pa.array([1, 2, 3], pa.int32())
    pa_array = pa.record_batch([pa_array_child], names=["some_column"])
    reader = pa.RecordBatchReader.from_batches(pa_array.schema, [pa_array])
    array_stream = na.c_array_stream(reader)

    arrays = list(array_stream)
    assert len(arrays) == 1
    assert arrays[0].schema.child(0).name == "some_column"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from nanoarrow._array_stream import CArrayStream
from nanoarrow._utils import obj_is_capsule
from nanoarrow.c_array import c_array
from nanoarrow.c_schema import c_schema


def c_array_stream(obj=None, schema=None) -> CArrayStream:
    """ArrowArrayStream wrapper

    This class provides a user-facing interface to access the fields of
    an ArrowArrayStream as defined in the Arrow C Stream interface.
    These objects are usually created using `nanoarrow.c_array_stream()`.

    Examples
    --------

    >>> import pyarrow as pa
    >>> import nanoarrow as na
    >>> pa_column = pa.array([1, 2, 3], pa.int32())
    >>> pa_batch = pa.record_batch([pa_column], names=["col1"])
    >>> pa_reader = pa.RecordBatchReader.from_batches(pa_batch.schema, [pa_batch])
    >>> array_stream = na.c_array_stream(pa_reader)
    >>> array_stream.get_schema()
    <nanoarrow.c_schema.CSchema struct>
    - format: '+s'
    - name: ''
    - flags: 0
    - metadata: NULL
    - dictionary: NULL
    - children[1]:
      'col1': <nanoarrow.c_schema.CSchema int32>
        - format: 'i'
        - name: 'col1'
        - flags: 2
        - metadata: NULL
        - dictionary: NULL
        - children[0]:
    >>> array_stream.get_next().length
    3
    >>> array_stream.get_next() is None
    Traceback (most recent call last):
      ...
    StopIteration
    """

    if schema is not None:
        schema = c_schema(schema)

    if isinstance(obj, CArrayStream) and schema is None:
        return obj

    # Try capsule protocol
    if hasattr(obj, "__arrow_c_stream__"):
        schema_capsule = None if schema is None else schema.__arrow_c_schema__()
        return CArrayStream._import_from_c_capsule(
            obj.__arrow_c_stream__(requested_schema=schema_capsule)
        )

    # Try import of bare capsule
    if obj_is_capsule(obj, "arrow_array_stream"):
        if schema is not None:
            raise TypeError(
                "Can't import c_array_stream from capsule with requested schema"
            )
        return CArrayStream._import_from_c_capsule(obj)

    # Try _export_to_c for RecordBatchReader objects if pyarrow < 14.0
    if _obj_is_pyarrow_record_batch_reader(obj):
        out = CArrayStream.allocate()
        obj._export_to_c(out._addr())
        return out

    try:
        array = c_array(obj, schema=schema)
        return CArrayStream.from_c_arrays([array], array.schema, validate=False)
    except Exception as e:
        raise TypeError(
            f"An error occurred whilst converting {type(obj).__name__} "
            f"to nanoarrow.c_array_stream or nanoarrow.c_array: \n {e}"
        ) from e


def allocate_c_array_stream() -> CArrayStream:
    """Allocate an uninitialized ArrowArrayStream wrapper

    Examples
    --------

    >>> import pyarrow as pa
    >>> from nanoarrow.c_array_stream import allocate_c_array_stream
    >>> pa_column = pa.array([1, 2, 3], pa.int32())
    >>> pa_batch = pa.record_batch([pa_column], names=["col1"])
    >>> pa_reader = pa.RecordBatchReader.from_batches(pa_batch.schema, [pa_batch])
    >>> array_stream = allocate_c_array_stream()
    >>> pa_reader._export_to_c(array_stream._addr())
    """
    return CArrayStream.allocate()


def _obj_is_pyarrow_record_batch_reader(obj):
    obj_type = type(obj)
    if not obj_type.__module__.startswith("pyarrow"):
        return False

    if not obj_type.__name__.endswith("RecordBatchReader"):
        return False

    return hasattr(obj, "_export_to_c")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Any, Iterable, Literal, Tuple, Union

from nanoarrow._array import CArray, CArrayBuilder, CArrayView, CDeviceArray
from nanoarrow._buffer import CBuffer, CBufferBuilder, NoneAwareWrapperIterator
from nanoarrow._device import DEVICE_CPU, Device
from nanoarrow._schema import CSchema, CSchemaBuilder
from nanoarrow._utils import obj_is_buffer, obj_is_capsule
from nanoarrow.c_buffer import c_buffer
from nanoarrow.c_schema import c_schema, c_schema_view
from nanoarrow.extension import resolve_extension

from nanoarrow import _types


def c_array(obj, schema=None) -> CArray:
    """ArrowArray wrapper

    This class provides a user-facing interface to access the fields of an ArrowArray
    as defined in the Arrow C Data interface, holding an optional reference to a
    :class:`CSchema` that can be used to safely deserialize the content.

    These objects are created using :func:`c_array`, which accepts any array-like
    object according to the Arrow PyCapsule interface, Python buffer protocol,
    or iterable of Python objects.

    This Python wrapper allows access to array fields but does not automatically
    deserialize their content: use :func:`c_array_view` to validate and deserialize
    the content into a more easily inspectable object.

    Note that the :class:`CArray` objects returned by ``.child()`` hold strong
    references to the original ``ArrowArray`` to avoid copies while inspecting an
    imported structure.

    Parameters
    ----------
    obj : array-like
        An object supporting the Arrow PyCapsule interface, the Python buffer
        protocol, or an iterable of Python objects.
    schema : schema-like or None
        A schema-like object as sanitized by :func:`c_schema` or None. This value
        will be used to request a data type from ``obj``; however, the conversion
        is best-effort (i.e., the data type of the returned ``CArray`` may be
        different than ``schema``).

    Examples
    --------

    >>> import nanoarrow as na
    >>> # Create from iterable
    >>> array = na.c_array([1, 2, 3], na.int32())
    >>> # Create from Python buffer (e.g., numpy array)
    >>> import numpy as np
    >>> array = na.c_array(np.array([1, 2, 3]))
    >>> # Create from Arrow PyCapsule (e.g., pyarrow array)
    >>> import pyarrow as pa
    >>> array = na.c_array(pa.array([1, 2, 3]))
    >>> # Access array fields
    >>> array.length
    3
    >>> array.null_count
    0
    """

    if schema is not None:
        schema = c_schema(schema)

    if isinstance(obj, CArray) and schema is None:
        return obj

    # Try Arrow PyCapsule protocol
    if hasattr(obj, "__arrow_c_array__"):
        schema_capsule = None if schema is None else schema.__arrow_c_schema__()
        return CArray._import_from_c_capsule(
            *obj.__arrow_c_array__(requested_schema=schema_capsule)
        )

    # Try import of bare capsule
    if obj_is_capsule(obj, "arrow_array"):
        if schema is None:
            schema_capsule = CSchema.allocate()._capsule
        else:
            schema_capsule = schema.__arrow_c_schema__()

        return CArray._import_from_c_capsule(schema_capsule, obj)

    # Try _export_to_c for Array/RecordBatch objects if pyarrow < 14.0
    if _obj_is_pyarrow_array(obj):
        out = CArray.allocate(CSchema.allocate())
        obj._export_to_c(out._addr(), out.schema._addr())
        return out

    # Use the ArrayBuilder classes to handle various strategies for other
    # types of objects (e.g., iterable, pybuffer, empty).
    try:
        builder_cls = _resolve_builder(obj)
    except Exception as e:
        raise TypeError(
            f"Can't convert object of type {type(obj).__name__} "
            f"to nanoarrow.c_array: \n {e}"
        ) from e

    try:
        if schema is None:
            obj, schema = builder_cls.infer_schema(obj)

        builder = builder_cls(schema)
        return builder.build_c_array(obj)
    except Exception as e:
        raise ValueError(
            f"An error occurred whilst converting {type(obj).__name__} "
            f"to nanoarrow.c_array: \n {e}"
        ) from e


def _resolve_builder(obj):
    if _obj_is_empty(obj):
        return EmptyArrayBuilder

    if obj_is_buffer(obj):
        return ArrayFromPyBufferBuilder

    if _obj_is_iterable(obj):
        return ArrayFromIterableBuilder

    raise TypeError(
        f"Can't resolve ArrayBuilder for object of type {type(obj).__name__}"
    )


def allocate_c_array(schema=None) -> CArray:
    """Allocate an uninitialized ArrowArray

    Examples
    --------

    >>> import pyarrow as pa
    >>> from nanoarrow.c_array import allocate_c_array
    >>> array = allocate_c_array()
    >>> pa.array([1, 2, 3])._export_to_c(array._addr())
    """
    if schema is not None:
        schema = c_schema(schema)

    return CArray.allocate(CSchema.allocate() if schema is None else schema)


def c_array_view(obj, schema=None) -> CArrayView:
    """ArrowArrayView wrapper

    The ``ArrowArrayView`` is a nanoarrow C library structure that provides
    structured access to buffers addresses, buffer sizes, and buffer
    data types. The buffer data is usually propagated from an ArrowArray
    but can also be propagated from other types of objects (e.g., serialized
    IPC). The offset and length of this view are independent of its parent
    (i.e., this object can also represent a slice of its parent).

    Examples
    --------

    >>> import pyarrow as pa
    >>> import numpy as np
    >>> import nanoarrow as na
    >>> from nanoarrow.c_array import c_array_view
    >>>
    >>> array = na.c_array(pa.array(["one", "two", "three", None]))
    >>> array_view = c_array_view(array)
    >>> np.array(array_view.buffer(1))
    array([ 0,  3,  6, 11, 11], dtype=int32)
    >>> np.array(array_view.buffer(2))
    array([b'o', b'n', b'e', b't', b'w', b'o', b't', b'h', b'r', b'e', b'e'],
          dtype='|S1')
    """

    if isinstance(obj, CArrayView) and schema is None:
        return obj

    return c_array(obj, schema).view()


def c_array_from_buffers(
    schema,
    length: int,
    buffers: Iterable[Any],
    null_count: int = -1,
    offset: int = 0,
    children: Iterable[Any] = (),
    validation_level: Literal[None, "full", "default", "minimal", "none"] = None,
    move: bool = False,
    device: Union[Device, None] = None,
) -> CArray:
    """Create an ArrowArray wrapper from components

    Given a schema, build an ArrowArray buffer-wise. This allows almost any array
    to be assembled; however, requires some knowledge of the Arrow Columnar
    specification. This function will do its best to validate the sizes and
    content of buffers according to ``validation_level``; however, not all
    types of arrays can currently be validated when constructed in this way.

    Parameters
    ----------
    schema : schema-like
        The data type of the desired array as sanitized by :func:`c_schema`.
    length : int
        The length of the output array.
    buffers : Iterable of buffer-like or None
        An iterable of buffers as sanitized by :func:`c_buffer`. Any object
        supporting the Python Buffer protocol is accepted. Buffer data types
        are not checked. A buffer value of ``None`` will skip setting a buffer
        (i.e., that buffer will be of length zero and its pointer will
        be ``NULL``).
    null_count : int, optional
        The number of null values, if known in advance. If -1 (the default),
        the null count will be calculated based on the validity bitmap. If
        the validity bitmap was set to ``None``, the calculated null count
        will be zero.
    offset : int, optional
        The logical offset from the start of the array.
    children : Iterable of array-like
        An iterable of arrays used to set child fields of the array. Can contain
        any object accepted by :func:`c_array`. Must contain the exact number of
        required children as specifed by ``schema``.
    validation_level: None or str, optional
        One of "none" (no check), "minimal" (check buffer sizes that do not require
        dereferencing buffer content), "default" (check all buffer sizes), or "full"
        (check all buffer sizes and all buffer content). The default, ``None``,
        will validate at the "default" level where possible.
    move : bool, optional
        Use ``True`` to move ownership of any input buffers or children to the
        output array.
    device : Device, optional
        An explicit device to use when constructing this array. If specified,
        this function will construct a :class:`CDeviceArray`; if unspecified,
        this function will construct a :class:`CArray` on the CPU device.

    Examples
    --------

    >>> import nanoarrow as na
    >>> c_array = na.c_array_from_buffers(na.uint8(), 5, [None, b"12345"])
    >>> na.Array(c_array).inspect()
    <ArrowArray uint8>
    - length: 5
    - offset: 0
    - null_count: 0
    - buffers[2]:
      - validity <bool[0 b] >
      - data <uint8[5 b] 49 50 51 52 53>
    - dictionary: NULL
    - children[0]:
    """
    if device is None:
        explicit_device = False
        device = DEVICE_CPU
    else:
        explicit_device = True

    schema = c_schema(schema)
    builder = CArrayBuilder.allocate(device)

    # Ensures that the output array->n_buffers is set and that the correct number
    # of children have been initialized.
    builder.init_from_schema(schema)

    # We might need more buffers if we're constructing a string or binary view
    buffers = list(buffers)
    builder.ensure_buffers(len(buffers))

    # Set buffers, optionally moving ownership of the buffers as well (i.e.,
    # the objects in the input buffers would be replaced with an empty ArrowBuffer)
    for i, buffer in enumerate(buffers):
        if buffer is None:
            continue

        # If we're setting a CBuffer from something else, we can avoid an extra
        # level of Python wrapping by using move=True
        move = move or not isinstance(buffer, CBuffer)
        builder.set_buffer(i, c_buffer(buffer), move=move)

    # Set children, optionally moving ownership of the children as well (i.e.,
    # the objects in the input children would be marked released).
    n_children = 0
    for child_src in children:
        # If we're setting a CArray from something else, we can avoid an extra
        # level of Python wrapping by using move=True
        move = move or not isinstance(child_src, (CArray, CDeviceArray))
        if move and isinstance(child_src, CDeviceArray):
            child_src = child_src.array

        builder.set_child(n_children, c_array(child_src), move=move)
        n_children += 1

    if n_children != schema.n_children:
        raise ValueError(f"Expected {schema.n_children} children but got {n_children}")

    # Set array fields
    builder.set_length(length)
    builder.set_offset(offset)
    builder.set_null_count(null_count)

    # Calculates the null count if -1 (and if applicable)
    builder.resolve_null_count()

    # Validate + finish. If device is specified (even CPU), always
    # return a device array.
    if not explicit_device:
        return builder.finish(validation_level=validation_level)
    else:
        return builder.finish_device()


class ArrayBuilder:
    """Internal utility to build CArrays from various types of input

    This class and its subclasses are designed to help separate the code
    that actually builds a CArray from the code that chooses the strategy
    used to do the building.
    """

    @classmethod
    def infer_schema(cls, obj) -> Tuple[CSchema, Any]:
        """Infer the Arrow data type from a target object

        Returns the type as a :class:`CSchema` and an object that can be
        consumed in the same way by append() in the event it had to be
        modified to infer its type (e.g., for an iterable, it would be
        necessary to consume the first element from the original iterator).
        """
        raise NotImplementedError()

    def __init__(self, schema):
        self._schema = c_schema(schema)
        self._schema_view = c_schema_view(self._schema)
        self._c_builder = CArrayBuilder.allocate()
        self._c_builder.init_from_schema(self._schema)

    def build_c_array(self, obj):
        self.start_building()
        self.append(obj)
        return self.finish_building()

    def start_building(self) -> None:
        pass

    def append(self, obj: Any) -> None:
        raise NotImplementedError()

    def finish_building(self) -> CArray:
        return self._c_builder.finish()


class EmptyArrayBuilder(ArrayBuilder):
    """Build an empty CArray of any type

    This builder accepts any empty input and produces a valid length zero
    array as output.
    """

    @classmethod
    def infer_schema(cls, obj) -> Tuple[Any, CSchema]:
        return obj, CSchemaBuilder.allocate().set_type(_types.NA)

    def start_building(self) -> None:
        self._c_builder.start_appending()

    def append(self, obj: Any) -> None:
        if len(obj) != 0:
            raise ValueError(
                f"Can't build empty array from {type(obj).__name__} "
                f"with length {len(obj)}"
            )


class ArrayFromPyBufferBuilder(ArrayBuilder):
    """Build a CArray from a Python Buffer

    This builder converts a Python buffer (e.g., numpy array, bytes, array.array)
    to a CArray (without copying the contents of the buffer).
    """

    @classmethod
    def infer_schema(cls, obj) -> Tuple[CBuffer, CSchema]:
        if not isinstance(obj, CBuffer):
            obj = CBuffer.from_pybuffer(obj)

        type_id = obj.data_type_id
        element_size_bits = obj.element_size_bits

        # Fixed-size binary needs a schema
        if type_id == _types.BINARY and element_size_bits != 0:
            schema = (
                CSchemaBuilder.allocate()
                .set_type_fixed_size(_types.FIXED_SIZE_BINARY, element_size_bits // 8)
                .finish()
            )
        elif type_id == _types.STRING:
            schema = CSchemaBuilder.allocate().set_type(_types.INT8).finish()
        elif type_id == _types.BINARY:
            schema = CSchemaBuilder.allocate().set_type(_types.UINT8).finish()
        else:
            schema = CSchemaBuilder.allocate().set_type(type_id).finish()

        return obj, schema

    def __init__(self, schema):
        super().__init__(schema)

        ext = resolve_extension(self._schema_view)
        self._append_ext = None
        if ext is not None:
            self._append_ext = ext.get_buffer_appender(self._schema, self)
        elif self._schema_view.extension_name:
            raise NotImplementedError(
                "Can't create array for unregistered extension "
                f"'{self._schema_view.extension_name}'"
            )

        if self._schema_view.storage_buffer_format is None:
            raise ValueError(
                f"Can't build array of type {self._schema_view.type} from PyBuffer"
            )

    def append(self, obj: Any) -> None:
        if not self._c_builder.is_empty():
            raise ValueError("Can't append to non-empty ArrayFromPyBufferBuilder")

        if not isinstance(obj, CBuffer):
            obj = CBuffer.from_pybuffer(obj)

        if self._append_ext is not None:
            return self._append_ext(obj)

        return self._append_impl(obj)

    def _append_impl(self, obj):
        if (
            self._schema_view.buffer_format in ("b", "c")
            and obj.format not in ("b", "c")
        ) and self._schema_view.buffer_format != obj.format:
            raise ValueError(
                f"Expected buffer with format '{self._schema_view.buffer_format}' "
                f"but got buffer with format '{obj.format}'"
            )

        self._c_builder.set_buffer(1, obj)
        self._c_builder.set_length(len(obj))
        self._c_builder.set_null_count(0)
        self._c_builder.set_offset(0)


class ArrayFromIterableBuilder(ArrayBuilder):
    """Build a CArray from an iterable of scalar objects

    This builder converts an iterable to a CArray using some heuristics to pick
    the fastest available method for converting to a particular type of array.
    Briefly, the methods are (1) ArrowArrayAppendXXX() functions from the C
    library (string, binary), (2) array.array() (integer/float except float16),
    (3) CBufferBuilder.write_elements() (everything else).
    """

    @classmethod
    def infer_schema(cls, obj) -> Tuple[CBuffer, CSchema]:
        raise ValueError("schema is required to build array from iterable")

    def __init__(self, schema):
        super().__init__(schema)

        # Resolve the method name we are going to use to do the building from
        # the provided schema.
        ext = resolve_extension(self._schema_view)
        if ext is not None:
            maybe_appender = ext.get_iterable_appender(self._schema, self)
            if maybe_appender:
                self._append_impl = maybe_appender
                return
        elif self._schema_view.extension_name:
            raise NotImplementedError(
                f"Can't create array for unregistered extension "
                f"'{self._schema_view.extension_name}'"
            )

        type_id = self._schema_view.type_id
        if type_id not in _ARRAY_BUILDER_FROM_ITERABLE_METHOD:
            raise ValueError(
                f"Can't build array of type {self._schema_view.type} from iterable"
            )

        method_name = _ARRAY_BUILDER_FROM_ITERABLE_METHOD[type_id]

        # If there might be nulls, we may need to pick a different strategy
        if (
            self._schema_view.nullable
            and method_name in _ARRAY_BUILDER_FROM_NULLABLE_ITERABLE_METHOD
        ):
            method_name = _ARRAY_BUILDER_FROM_NULLABLE_ITERABLE_METHOD[method_name]

        self._append_impl = getattr(self, method_name)

    def start_building(self) -> None:
        self._c_builder.start_appending()

    def append(self, obj: Any) -> None:
        self._append_impl(obj)

    def _append_strings(self, obj: Iterable) -> None:
        self._c_builder.append_strings(obj)

    def _append_bytes(self, obj: Iterable) -> None:
        self._c_builder.append_bytes(obj)

    def _build_nullable_array_using_array(self, obj: Iterable) -> None:
        wrapper = NoneAwareWrapperIterator(
            obj, self._schema_view.storage_type_id, self._schema_view.fixed_size
        )
        self._append_using_array(wrapper)

        _, null_count, validity = wrapper.finish()
        if validity is not None:
            self._c_builder.set_buffer(0, validity, move=True)

        self._c_builder.set_null_count(null_count)

    def _build_nullable_array_using_buffer_builder(self, obj: Iterable) -> None:
        wrapper = NoneAwareWrapperIterator(
            obj, self._schema_view.storage_type_id, self._schema_view.fixed_size
        )
        self._append_using_buffer_builder(wrapper)

        _, null_count, validity = wrapper.finish()
        if validity is not None:
            self._c_builder.set_buffer(0, validity, move=True)

        self._c_builder.set_null_count(null_count)

    def _append_using_array(self, obj: Iterable) -> None:
        from array import array

        py_array = array(self._schema_view.storage_buffer_format, obj)
        buffer = CBuffer.from_pybuffer(py_array)
        self._c_builder.set_buffer(1, buffer, move=True)
        self._c_builder.set_length(len(buffer))
        self._c_builder.set_null_count(0)
        self._c_builder.set_offset(0)

    def _append_using_buffer_builder(self, obj: Iterable) -> None:
        builder = CBufferBuilder()
        builder.set_data_type(self._schema_view.type_id)

        n_values = builder.write_elements(obj)

        buffer = builder.finish()
        self._c_builder.set_buffer(1, buffer, move=True)
        self._c_builder.set_length(n_values)
        self._c_builder.set_null_count(0)
        self._c_builder.set_offset(0)


_ARRAY_BUILDER_FROM_ITERABLE_METHOD = {
    _types.BOOL: "_append_using_buffer_builder",
    _types.HALF_FLOAT: "_append_using_buffer_builder",
    _types.INTERVAL_MONTH_DAY_NANO: "_append_using_buffer_builder",
    _types.INTERVAL_DAY_TIME: "_append_using_buffer_builder",
    _types.INTERVAL_MONTHS: "_append_using_buffer_builder",
    _types.BINARY: "_append_bytes",
    _types.LARGE_BINARY: "_append_bytes",
    _types.FIXED_SIZE_BINARY: "_append_bytes",
    _types.BINARY_VIEW: "_append_bytes",
    _types.STRING: "_append_strings",
    _types.LARGE_STRING: "_append_strings",
    _types.STRING_VIEW: "_append_strings",
    _types.INT8: "_append_using_array",
    _types.UINT8: "_append_using_array",
    _types.INT16: "_append_using_array",
    _types.UINT16: "_append_using_array",
    _types.INT32: "_append_using_array",
    _types.UINT32: "_append_using_array",
    _types.INT64: "_append_using_array",
    _types.UINT64: "_append_using_array",
    _types.FLOAT: "_append_using_array",
    _types.DOUBLE: "_append_using_array",
    _types.TIMESTAMP: "_append_using_array",
    _types.DATE32: "_append_using_array",
    _types.DATE64: "_append_using_array",
    _types.DURATION: "_append_using_array",
}

_ARRAY_BUILDER_FROM_NULLABLE_ITERABLE_METHOD = {
    "_append_using_array": "_build_nullable_array_using_array",
    "_append_using_buffer_builder": "_build_nullable_array_using_buffer_builder",
}


def _obj_is_iterable(obj):
    return hasattr(obj, "__iter__")


def _obj_is_empty(obj):
    return hasattr(obj, "__len__") and len(obj) == 0


# This is a heuristic for detecting a pyarrow.Array or pyarrow.RecordBatch
# for pyarrow < 14.0.0, after which the the __arrow_c_array__ protocol
# is sufficient to detect such an array. This check can't use isinstance()
# to avoid importing pyarrow unnecessarily.
def _obj_is_pyarrow_array(obj):
    obj_type = type(obj)
    if not obj_type.__module__.startswith("pyarrow"):
        return False

    if not obj_type.__name__.endswith("Array") and obj_type.__name__ != "RecordBatch":
        return False

    return hasattr(obj, "_export_to_c")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import io
import os
import pathlib
import tempfile

import pytest
from nanoarrow._utils import NanoarrowException

import nanoarrow as na
from nanoarrow.ipc import _EXAMPLE_IPC_SCHEMA, InputStream, StreamWriter


def test_ipc_stream_example():
    with InputStream.example() as input:
        assert input._is_valid() is True
        assert "BytesIO object" in repr(input)

        stream = na.c_array_stream(input)
        assert input._is_valid() is False
        assert stream.is_valid() is True
        assert repr(input) == "<nanoarrow.ipc.InputStream <invalid>>"
        with pytest.raises(RuntimeError, match="no longer valid"):
            stream = na.c_array_stream(input)

        with stream:
            schema = stream.get_schema()
            assert schema.format == "+s"
            assert schema.child(0).format == "i"
            batches = list(stream)
            assert stream.is_valid() is True

        assert stream.is_valid() is False
        assert len(batches) == 1
        batch = na.c_array(batches[0]).view()
        assert list(batch.child(0).buffer(1)) == [1, 2, 3]


def test_ipc_stream_from_readable():
    with io.BytesIO(InputStream.example_bytes()) as f:
        with InputStream.from_readable(f) as input:
            assert input._is_valid() is True
            assert "BytesIO object" in repr(input)

            with na.c_array_stream(input) as stream:
                batches = list(stream)
                assert len(batches) == 1
                assert len(batches[0]) == 3


def test_ipc_stream_from_path():
    with tempfile.TemporaryDirectory() as td:
        path = os.path.join(td, "test.arrows")
        with open(path, "wb") as f:
            f.write(InputStream.example_bytes())

        with InputStream.from_path(path) as input:
            assert repr(path) in repr(input)
            with na.c_array_stream(input) as stream:
                batches = list(stream)
                assert len(batches) == 1
                assert len(batches[0]) == 3


def test_ipc_stream_from_url():
    with tempfile.TemporaryDirectory() as td:
        path = os.path.join(td, "test.arrows")
        with open(path, "wb") as f:
            f.write(InputStream.example_bytes())

        uri = pathlib.Path(path).as_uri()
        with InputStream.from_url(uri) as input:
            with na.c_array_stream(input) as stream:
                batches = list(stream)
                assert len(batches) == 1
                assert len(batches[0]) == 3


def test_ipc_stream_compressed_example():
    buf = io.BytesIO()
    buf.write(_EXAMPLE_IPC_SCHEMA)
    buf.write(COMPRESSED_BATCH)
    buf.seek(0)

    with InputStream.from_readable(buf) as inp:
        array = na.Array(inp)
        assert len(array) == 3
        assert array.child(0).to_pylist() == [0, 1, 2]


def test_ipc_stream_python_exception_on_read():
    class ExtraordinarilyInconvenientFile:
        def readinto(self, *args, **kwargs):
            raise RuntimeError("I error for all read requests")

    input = InputStream.from_readable(ExtraordinarilyInconvenientFile())
    with pytest.raises(
        NanoarrowException, match="RuntimeError: I error for all read requests"
    ):
        na.c_array_stream(input)


def test_ipc_stream_error_on_read():
    with io.BytesIO(InputStream.example_bytes()[:100]) as f:
        with InputStream.from_readable(f) as input:

            with pytest.raises(
                NanoarrowException,
                match="Expected >= 280 bytes of remaining data",
            ):
                na.c_array_stream(input)


def test_writer_from_writable():
    array = na.c_array_from_buffers(
        na.struct({"some_col": na.int32()}),
        length=3,
        buffers=[],
        children=[na.c_array([1, 2, 3], na.int32())],
    )

    out = io.BytesIO()
    with StreamWriter.from_writable(out) as writer:
        writer.write_array(array)

    with na.ArrayStream.from_readable(out.getvalue()) as stream:
        assert stream.read_all().to_pylist() == na.Array(array).to_pylist()


def test_writer_from_path():
    array = na.c_array_from_buffers(
        na.struct({"some_col": na.int32()}),
        length=3,
        buffers=[],
        children=[na.c_array([1, 2, 3], na.int32())],
    )

    with tempfile.TemporaryDirectory() as td:
        path = os.path.join(td, "test.arrows")

        with StreamWriter.from_path(path) as writer:
            writer.write_array(array)

        with na.ArrayStream.from_path(path) as stream:
            assert stream.read_all().to_pylist() == na.Array(array).to_pylist()


def test_writer_write_stream_schema():
    array = na.c_array_from_buffers(
        na.struct({"some_col": na.int32()}),
        length=3,
        buffers=[],
        children=[na.c_array([1, 2, 3], na.int32())],
    )
    zero_chunk_array = na.Array.from_chunks([], array.schema)

    out = io.BytesIO()
    with StreamWriter.from_writable(out) as writer:
        writer.write_stream(zero_chunk_array)
        schema_bytes = out.getvalue()

    with StreamWriter.from_writable(out) as writer:
        out.truncate(0)
        out.seek(0)
        writer.write_stream(zero_chunk_array)
        writer.write_stream(zero_chunk_array, write_schema=True)
        two_schema_bytes = out.getvalue()

    assert (schema_bytes + schema_bytes) == two_schema_bytes

    with StreamWriter.from_writable(out) as writer:
        out.truncate(0)
        out.seek(0)
        writer.write_array(array, write_schema=False)
        array_bytes = out.getvalue()

    with StreamWriter.from_writable(out) as writer:
        out.truncate(0)
        out.seek(0)
        writer.write_array(array, write_schema=True)
        both_bytes = out.getvalue()

    assert (schema_bytes + array_bytes) == both_bytes


def test_writer_serialize_stream():
    array = na.c_array_from_buffers(
        na.struct({"some_col": na.int32()}),
        length=3,
        buffers=[],
        children=[na.c_array([1, 2, 3], na.int32())],
    )

    out = io.BytesIO()
    with StreamWriter.from_writable(out) as writer:
        writer.write_array(array)

        # Check that we can't serialize after we've already written to stream
        with pytest.raises(ValueError, match="Can't serialize_stream"):
            writer.serialize_stream(array)

        schema_and_array_bytes = out.getvalue()

    end_of_stream_bytes = b"\xff\xff\xff\xff\x00\x00\x00\x00"

    writer = StreamWriter.from_writable(out)
    out.truncate(0)
    out.seek(0)
    writer.serialize_stream(array)
    assert writer._is_valid() is False

    serialized_bytes = out.getvalue()
    assert (schema_and_array_bytes + end_of_stream_bytes) == serialized_bytes


def test_writer_python_exception_on_write():
    class ExtraordinarilyInconvenientFile:
        def write(self, *args, **kwargs):
            raise RuntimeError("I error for all write requests")

    with pytest.raises(NanoarrowException, match="I error for all write requests"):
        with StreamWriter.from_writable(ExtraordinarilyInconvenientFile()) as writer:
            writer.write(na.c_array([], na.struct([na.int32()])))


def test_writer_error_on_write():
    with pytest.raises(NanoarrowException):
        with StreamWriter.from_writable(io.BytesIO()) as writer:
            writer.write_stream(na.c_array([], na.int32()))


# fmt: off
COMPRESSED_BATCH = bytearray([
    0xff, 0xff, 0xff, 0xff, 0xa0, 0x00, 0x00, 0x00, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x0c, 0x00, 0x18, 0x00, 0x06, 0x00, 0x05, 0x00, 0x08, 0x00, 0x0c, 0x00,
    0x0c, 0x00, 0x00, 0x00, 0x00, 0x03, 0x04, 0x00, 0x1c, 0x00, 0x00, 0x00, 0x20, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0c, 0x00, 0x1e, 0x00,
    0x10, 0x00, 0x04, 0x00, 0x08, 0x00, 0x0c, 0x00, 0x0c, 0x00, 0x00, 0x00, 0x50, 0x00,
    0x00, 0x00, 0x24, 0x00, 0x00, 0x00, 0x18, 0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x00, 0x08, 0x00,
    0x07, 0x00, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1d, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x03, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x0c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x28, 0xb5, 0x2f, 0xfd, 0x20, 0x0c,
    0x61, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00
])
# fmt: on



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import io
import os
import pathlib
import re


def read_content(path_or_content):
    if isinstance(path_or_content, pathlib.Path):
        with open(path_or_content) as f:
            return f.read()
    else:
        return str(path_or_content)


def write_content(path_or_content, out_path):
    with open(out_path, "w") as f:
        f.write(read_content(path_or_content))


def configure_content(paths_or_content, args):
    content = read_content(paths_or_content)

    for key, value in args.items():
        replace_key = f"@{key}@"
        if content.count(replace_key) != 1:
            raise ValueError(
                "Expected exactly one occurrence of "
                f"'{replace_key}' in '{paths_or_content}'"
            )

        content = content.replace(replace_key, str(value))

    return content


def concatenate_content(paths_or_content):
    out = io.StringIO()

    for path in paths_or_content:
        out.write(read_content(path))

    return out.getvalue()


def cmakelist_version(path_or_content):
    content = read_content(path_or_content)

    version_match = re.search(r'set\(NANOARROW_VERSION "(.*?)"\)', content)
    if version_match is None:
        raise ValueError(f"Can't find NANOARROW_VERSION in '{path_or_content}'")

    version = version_match.group(1)

    component_match = re.search(r"^([0-9]+)\.([0-9]+)\.([0-9]+)", version)
    return (version,) + tuple(int(component) for component in component_match.groups())


def namespace_nanoarrow_includes(path_or_content, header_namespace="nanoarrow"):
    content = read_content(path_or_content)
    return re.sub(
        r'#include "nanoarrow/([^"]+)"', f'#include "{header_namespace}\\1"', content
    )


def bundle_nanoarrow(
    root_dir,
    symbol_namespace=None,
    header_namespace="nanoarrow/",
    output_source_dir="src",
    output_include_dir="include",
    cpp=False,
):
    root_dir = pathlib.Path(root_dir)
    src_dir = root_dir / "src" / "nanoarrow"

    output_source_dir = pathlib.Path(output_source_dir)
    output_include_dir = pathlib.Path(output_include_dir) / header_namespace

    version, major, minor, patch = cmakelist_version(root_dir / "CMakeLists.txt")

    if symbol_namespace is None:
        namespace_define = "// #define NANOARROW_NAMESPACE YourNamespaceHere"
    else:
        namespace_define = f"#define NANOARROW_NAMESPACE {symbol_namespace}"

    nanoarrow_config_h = configure_content(
        src_dir / "nanoarrow_config.h.in",
        {
            "NANOARROW_VERSION": version,
            "NANOARROW_VERSION_MAJOR": major,
            "NANOARROW_VERSION_MINOR": minor,
            "NANOARROW_VERSION_PATCH": patch,
            "NANOARROW_NAMESPACE_DEFINE": namespace_define,
        },
    )

    # Generate nanoarrow/nanoarrow.h
    nanoarrow_h = concatenate_content(
        [
            nanoarrow_config_h,
            src_dir / "common" / "inline_types.h",
            src_dir / "nanoarrow.h",
            src_dir / "common" / "inline_buffer.h",
            src_dir / "common" / "inline_array.h",
        ]
    )

    nanoarrow_h = re.sub(r'#include "(nanoarrow/)?[a-z_./]+"', "", nanoarrow_h)
    yield f"{output_include_dir}/nanoarrow.h", nanoarrow_h

    # Generate nanoarrow/nanoarrow.hpp
    nanoarrow_hpp = concatenate_content(
        [
            src_dir / "nanoarrow.hpp",
            src_dir / "hpp" / "exception.hpp",
            src_dir / "hpp" / "operators.hpp",
            src_dir / "hpp" / "unique.hpp",
            src_dir / "hpp" / "array_stream.hpp",
            src_dir / "hpp" / "buffer.hpp",
            src_dir / "hpp" / "view.hpp",
        ]
    )

    nanoarrow_hpp = re.sub(r'#include "(nanoarrow/)?hpp/[a-z_./]+"', "", nanoarrow_hpp)
    nanoarrow_hpp = namespace_nanoarrow_includes(nanoarrow_hpp, header_namespace)
    yield f"{output_include_dir}/nanoarrow.hpp", nanoarrow_hpp

    # Generate nanoarrow/nanoarrow.c
    nanoarrow_c = concatenate_content(
        [
            src_dir / "common" / "utils.c",
            src_dir / "common" / "schema.c",
            src_dir / "common" / "array.c",
            src_dir / "common" / "array_stream.c",
        ]
    )
    nanoarrow_c = namespace_nanoarrow_includes(nanoarrow_c, header_namespace)

    if cpp:
        yield f"{output_source_dir}/nanoarrow.cc", nanoarrow_c
    else:
        yield f"{output_source_dir}/nanoarrow.c", nanoarrow_c


def bundle_nanoarrow_device(
    root_dir,
    header_namespace="nanoarrow/",
    output_source_dir="src",
    output_include_dir="include",
):
    root_dir = pathlib.Path(root_dir)
    src_dir = root_dir / "src" / "nanoarrow"

    output_source_dir = pathlib.Path(output_source_dir)
    output_include_dir = pathlib.Path(output_include_dir) / header_namespace

    # Generate headers
    for filename in ["nanoarrow_device.h", "nanoarrow_device.hpp"]:
        content = read_content(src_dir / filename)
        content = namespace_nanoarrow_includes(content, header_namespace)
        yield f"{output_include_dir}/{filename}", content

    # Generate sources
    content = concatenate_content(
        [src_dir / "device" / "device.c", src_dir / "device" / "cuda.c"]
    )
    content = namespace_nanoarrow_includes(content, header_namespace)
    yield f"{output_source_dir}/nanoarrow_device.c", content


def bundle_nanoarrow_ipc(
    root_dir,
    header_namespace="nanoarrow/",
    output_source_dir="src",
    output_include_dir="include",
):
    root_dir = pathlib.Path(root_dir)
    src_dir = root_dir / "src" / "nanoarrow"

    output_source_dir = pathlib.Path(output_source_dir)
    output_include_dir = pathlib.Path(output_include_dir) / header_namespace

    # Generate headers
    for filename in [
        "nanoarrow_ipc.h",
        "nanoarrow_ipc.hpp",
    ]:
        content = read_content(src_dir / filename)
        content = namespace_nanoarrow_includes(content, header_namespace)
        yield f"{output_include_dir}/{filename}", content

    nanoarrow_ipc_c = concatenate_content(
        [
            src_dir / "ipc" / "flatcc_generated.h",
            src_dir / "ipc" / "codecs.c",
            src_dir / "ipc" / "decoder.c",
            src_dir / "ipc" / "encoder.c",
            src_dir / "ipc" / "reader.c",
            src_dir / "ipc" / "writer.c",
        ]
    )
    nanoarrow_ipc_c = nanoarrow_ipc_c.replace(
        '#include "nanoarrow/ipc/flatcc_generated.h"', ""
    )
    nanoarrow_ipc_c = namespace_nanoarrow_includes(nanoarrow_ipc_c, header_namespace)
    yield f"{output_source_dir}/nanoarrow_ipc.c", nanoarrow_ipc_c


def bundle_nanoarrow_testing(
    root_dir,
    header_namespace="nanoarrow/",
    output_source_dir="src",
    output_include_dir="include",
):
    root_dir = pathlib.Path(root_dir)
    src_dir = root_dir / "src" / "nanoarrow"

    output_source_dir = pathlib.Path(output_source_dir)
    output_include_dir = pathlib.Path(output_include_dir) / header_namespace

    # Generate headers
    for filename in [
        "nanoarrow_testing.hpp",
        "nanoarrow_gtest_util.hpp",
    ]:
        content = read_content(src_dir / filename)
        content = namespace_nanoarrow_includes(content, header_namespace)
        yield f"{output_include_dir}/{filename}", content

    nanoarrow_testing_cc = concatenate_content(
        [
            src_dir / "testing" / "testing.cc",
        ]
    )
    nanoarrow_testing_cc = namespace_nanoarrow_includes(
        nanoarrow_testing_cc, header_namespace
    )
    yield f"{output_source_dir}/nanoarrow_testing.cc", nanoarrow_testing_cc


def bundle_flatcc(
    root_dir,
    output_source_dir="src",
    output_include_dir="include",
):
    root_dir = pathlib.Path(root_dir)
    flatcc_dir = root_dir / "thirdparty" / "flatcc"

    output_source_dir = pathlib.Path(output_source_dir)
    output_include_dir = pathlib.Path(output_include_dir)

    # Generate headers
    include_dir = flatcc_dir / "include"
    for abs_filename in include_dir.glob("flatcc/**/*.h"):
        filename = abs_filename.relative_to(include_dir)
        yield f"{output_include_dir}/{filename}", read_content(
            flatcc_dir / "include" / filename
        )

    # Generate sources
    src_dir = flatcc_dir / "src" / "runtime"
    flatcc_c = concatenate_content(
        [
            src_dir / "builder.c",
            src_dir / "emitter.c",
            src_dir / "verifier.c",
            src_dir / "refmap.c",
        ]
    )

    yield f"{output_source_dir}/flatcc.c", flatcc_c


def ensure_output_path_exists(out_path: pathlib.Path):
    if out_path.is_dir() and out_path.exists():
        return

    if out_path.is_file() and out_path.exists():
        raise ValueError(f"Can't create directory '{out_path}': exists and is a file")

    ensure_output_path_exists(out_path.parent)
    os.mkdir(out_path)


def do_bundle(bundler):
    for out_file, out_content in bundler:
        out_path = pathlib.Path(out_file)
        ensure_output_path_exists(out_path.parent)
        write_content(out_content, out_path)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Bundled nanoarrow distribution")
    parser.add_argument(
        "--include-output-dir",
        help="include/ directory in which nanoarrow headers should be placed",
    )
    parser.add_argument(
        "--source-output-dir",
        help="Directory in which nanoarrow source files should be placed",
    )
    parser.add_argument(
        "--symbol-namespace", help="A value with which symbols should be prefixed"
    )
    parser.add_argument(
        "--header-namespace",
        help=(
            "The directory within include-output-dir that nanoarrow headers should be"
            "placed"
        ),
        default="nanoarrow/",
    )
    parser.add_argument(
        "--output-dir",
        help=(
            "If include-output-dir or source-output-dir are missing, ensures a single "
            "output directory with include/ and src/ subdirectories containing the "
            "headers and sources, respectively"
        ),
        default="dist",
    )
    parser.add_argument(
        "--cpp", help="Bundle sources as C++ where possible", action="store_true"
    )
    parser.add_argument(
        "--with-device",
        help="Include nanoarrow_device sources/headers",
        action="store_true",
    )
    parser.add_argument(
        "--with-ipc",
        help="Include nanoarrow_ipc sources/headers",
        action="store_true",
    )
    parser.add_argument(
        "--with-testing",
        help="Include nanoarrow_testing sources/headers",
        action="store_true",
    )
    parser.add_argument(
        "--with-flatcc",
        help="Include flatcc sources/headers",
        action="store_true",
    )

    args = parser.parse_args()
    if args.include_output_dir is None:
        args.include_output_dir = pathlib.Path(args.output_dir) / "include"
    if args.source_output_dir is None:
        args.source_output_dir = pathlib.Path(args.output_dir) / "src"

    root_dir = pathlib.Path(__file__).parent.parent.parent

    # Bundle nanoarrow
    do_bundle(
        bundle_nanoarrow(
            root_dir,
            symbol_namespace=args.symbol_namespace,
            header_namespace=args.header_namespace,
            output_source_dir=args.source_output_dir,
            output_include_dir=args.include_output_dir,
            cpp=args.cpp,
        )
    )

    # Bundle nanoarrow_device
    if args.with_device:
        do_bundle(
            bundle_nanoarrow_device(
                root_dir,
                header_namespace=args.header_namespace,
                output_source_dir=args.source_output_dir,
                output_include_dir=args.include_output_dir,
            )
        )

    # Bundle nanoarrow_ipc
    if args.with_ipc:
        do_bundle(
            bundle_nanoarrow_ipc(
                root_dir,
                header_namespace=args.header_namespace,
                output_source_dir=args.source_output_dir,
                output_include_dir=args.include_output_dir,
            )
        )

    # Bundle nanoarrow_testing
    if args.with_testing:
        do_bundle(
            bundle_nanoarrow_testing(
                root_dir,
                header_namespace=args.header_namespace,
                output_source_dir=args.source_output_dir,
                output_include_dir=args.include_output_dir,
            )
        )

    # Bundle flatcc
    if args.with_flatcc:
        do_bundle(
            bundle_flatcc(
                root_dir,
                output_source_dir=args.source_output_dir,
                output_include_dir=args.include_output_dir,
            )
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import pytest
from nanoarrow.c_buffer import CBuffer

import nanoarrow as na
from nanoarrow import visitor


def test_to_pylist():
    array = na.c_array([1, 2, 3], na.int32())
    assert visitor.ToPyListConverter.visit(array) == [1, 2, 3]


def test_convert():
    ints = na.c_array([1, 2, 3], na.int32())
    bools = na.c_array([1, 0, 1], na.bool_())
    strings = na.c_array(["abc", "def", "ghi"], na.string())

    ints_col = visitor.ToPySequenceConverter.visit(ints)
    assert isinstance(ints_col, CBuffer)
    assert ints_col.format == "i"
    assert list(ints_col) == [1, 2, 3]

    bools_col = visitor.ToPySequenceConverter.visit(bools)
    assert isinstance(bools_col, CBuffer)
    assert bools_col.format == "?"
    assert list(bools_col) == [True, False, True]

    strings_col = visitor.ToPySequenceConverter.visit(strings)
    assert isinstance(strings_col, list)
    assert strings_col == ["abc", "def", "ghi"]


def test_convert_non_nullable():
    ints = na.c_array([1, 2, 3], na.int32(nullable=False))
    bools = na.c_array([1, 0, 1], na.bool_(nullable=False))
    strings = na.c_array(["abc", "def", "ghi"], na.string(nullable=False))

    ints_col = visitor.ToPySequenceConverter.visit(ints)
    assert isinstance(ints_col, CBuffer)
    assert ints_col.format == "i"
    assert list(ints_col) == [1, 2, 3]

    bools_col = visitor.ToPySequenceConverter.visit(bools)
    assert isinstance(bools_col, CBuffer)
    assert bools_col.format == "?"
    assert list(bools_col) == [True, False, True]

    strings_col = visitor.ToPySequenceConverter.visit(strings)
    assert isinstance(strings_col, list)
    assert strings_col == ["abc", "def", "ghi"]


def test_convert_columns():
    array = na.c_array_from_buffers(
        na.struct({"col1": na.int32(), "col2": na.bool_(), "col3": na.string()}),
        length=3,
        buffers=[None],
        children=[
            na.c_array([1, 2, 3], na.int32()),
            na.c_array([1, 0, 1], na.bool_()),
            na.c_array(["abc", "def", "ghi"], na.string()),
        ],
    )

    names, columns = visitor.ToColumnsPysequenceConverter.visit(array)
    assert names == ["col1", "col2", "col3"]
    assert list(columns[0]) == [1, 2, 3]
    assert list(columns[1]) == [True, False, True]
    assert columns[2] == ["abc", "def", "ghi"]

    with pytest.raises(ValueError, match="can only be used on a struct array"):
        visitor.ToColumnsPysequenceConverter.visit([], na.int32())

    # Ensure that the columns converter errors for top-level nulls
    array_with_nulls = na.c_array_from_buffers(
        array.schema,
        array.length,
        [na.c_buffer([True, False, True], na.bool_())],
        children=array.children,
    )
    with pytest.raises(ValueError, match="null_count > 0"):
        visitor.ToColumnsPysequenceConverter.visit(array_with_nulls)


def test_contiguous_buffer_converter():
    array = na.Array.from_chunks([[1, 2, 3], [4, 5, 6]], na.int32())
    buffer = visitor.ToPyBufferConverter.visit(array)
    assert list(buffer) == [1, 2, 3, 4, 5, 6]


def test_contiguous_buffer_converter_with_offsets():
    src = [na.c_array([1, 2, 3], na.int32())[1:], na.c_array([4, 5, 6], na.int32())[2:]]
    array = na.Array.from_chunks(src)
    buffer = visitor.ToPyBufferConverter.visit(array)
    assert list(buffer) == [2, 3, 6]


def test_boolean_bytes_converter():
    array = na.Array.from_chunks([[0, 1, 1], [1, 0, 0]], na.bool_())
    buffer = visitor.ToBooleanBufferConverter.visit(array)
    assert list(buffer) == [False, True, True, True, False, False]


def test_boolean_bytes_converter_with_offsets():
    src = [na.c_array([0, 1, 1], na.bool_())[1:], na.c_array([1, 0, 0], na.bool_())[2:]]
    array = na.Array.from_chunks(src)
    buffer = visitor.ToBooleanBufferConverter.visit(array)
    assert list(buffer) == [True, True, False]


def test_nullable_converter():
    # All valid
    array = na.Array.from_chunks([[1, 2, 3], [4, 5, 6]], na.int32())
    is_valid, column = visitor.ToNullableSequenceConverter.visit(
        array, handle_nulls=na.nulls_separate()
    )
    assert is_valid is None
    assert list(column) == [1, 2, 3, 4, 5, 6]

    # Only nulls in the first chunk
    array = na.Array.from_chunks([[1, None, 3], [4, 5, 6]], na.int32())
    is_valid, column = visitor.ToNullableSequenceConverter.visit(
        array, handle_nulls=na.nulls_separate()
    )
    assert list(is_valid) == [True, False, True, True, True, True]
    assert list(column) == [1, 0, 3, 4, 5, 6]

    # Only nulls in the second chunk
    array = na.Array.from_chunks([[1, 2, 3], [4, None, 6]], na.int32())
    is_valid, column = visitor.ToNullableSequenceConverter.visit(
        array, handle_nulls=na.nulls_separate()
    )
    assert list(is_valid) == [True, True, True, True, False, True]
    assert list(column) == [1, 2, 3, 4, 0, 6]

    # Nulls in both chunks
    array = na.Array.from_chunks([[1, None, 3], [4, None, 6]], na.int32())
    is_valid, column = visitor.ToNullableSequenceConverter.visit(
        array, handle_nulls=na.nulls_separate()
    )
    assert list(is_valid) == [True, False, True, True, False, True]
    assert list(column) == [1, 0, 3, 4, 0, 6]


def test_nulls_forbid():
    is_valid_non_empty = na.c_buffer([1, 0, 1], na.uint8())
    data = na.c_buffer([1, 2, 3], na.int32())

    forbid_nulls = visitor.nulls_forbid()
    assert forbid_nulls(None, data) is data
    with pytest.raises(ValueError):
        forbid_nulls(is_valid_non_empty, data)


def test_numpy_null_handling():
    np = pytest.importorskip("numpy")

    is_valid_non_empty = memoryview(na.c_buffer([1, 0, 1], na.uint8())).cast("?")
    data = na.c_buffer([1, 2, 3], na.int32())

    # Check nulls as sentinel
    nulls_as_sentinel = visitor.nulls_as_sentinel()
    np.testing.assert_array_equal(
        nulls_as_sentinel(None, data), np.array([1, 2, 3], np.int32)
    )
    np.testing.assert_array_equal(
        nulls_as_sentinel(is_valid_non_empty, data),
        np.array([1, np.nan, 3], dtype=np.float64),
    )


def test_iterator_unregistered_extension():
    from nanoarrow.iterator import UnregisteredExtensionWarning

    schema = na.extension_type(na.int32(), "arrow.test")
    storage_array = na.c_array([1, 2, 3], na.int32())
    _, storage_array_capsule = na.c_array(storage_array).__arrow_c_array__()
    extension_array = na.c_array(storage_array_capsule, schema)

    with pytest.warns(UnregisteredExtensionWarning):
        visitor.ToPyListConverter.visit(extension_array)

    with pytest.warns(UnregisteredExtensionWarning):
        visitor.ToPySequenceConverter.visit(extension_array)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import _cython_3_0_11
from _typeshed import Incomplete
from typing import ClassVar

__pyx_capi__: dict
__reduce_cython__: _cython_3_0_11.cython_function_or_method
__setstate_cython__: _cython_3_0_11.cython_function_or_method
__test__: dict
assert_type_equal: _cython_3_0_11.cython_function_or_method

class CArrowTimeUnit:
    MICRO: ClassVar[int] = ...
    MILLI: ClassVar[int] = ...
    NANO: ClassVar[int] = ...
    SECOND: ClassVar[int] = ...
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    def __reduce__(self): ...
    def __reduce_cython__(self, *args, **kwargs): ...
    def __setstate_cython__(self, *args, **kwargs): ...

class CLayout:
    buffer_data_type_id: Incomplete
    child_size_elements: Incomplete
    element_size_bits: Incomplete
    n_buffers: Incomplete
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    def __reduce__(self): ...

class CSchema:
    children: Incomplete
    dictionary: Incomplete
    flags: Incomplete
    format: Incomplete
    metadata: Incomplete
    n_children: Incomplete
    name: Incomplete
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    @staticmethod
    def allocate(*args, **kwargs):
        """Allocate a released CSchema"""
    def child(self, *args, **kwargs): ...
    def is_valid(self, *args, **kwargs):
        """Check for a non-null and non-released underlying ArrowSchema"""
    def modify(self, *args, **kwargs): ...
    def type_equals(self, *args, **kwargs):
        """Test two schemas for data type equality

        Checks two CSchema objects for type equality (i.e., that an array with
        schema ``actual`` contains elements with the same logical meaning as and
        array with schema ``expected``). Notably, this excludes metadata from
        all nodes in the schema.

        Parameters
        ----------
        other : CSchema
            The schema against which to test
        check_nullability : bool
            If True, actual and expected will be considered equal if their
            data type information and marked nullability are identical.
        """
    def __arrow_c_schema__(self, *args, **kwargs):
        """
        Export to a ArrowSchema PyCapsule
        """
    def __deepcopy__(self): ...
    def __reduce__(self): ...

class CSchemaBuilder:
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    @staticmethod
    def allocate(*args, **kwargs):
        """Create a CSchemaBuilder

        Allocates memory for an ArrowSchema and populates it with nanoarrow's
        ArrowSchema private_data/release callback implementation. This should
        usually be followed by :meth:`set_type` or :meth:`set_format`.
        """
    def allocate_children(self, *args, **kwargs): ...
    def append_metadata(self, *args, **kwargs):
        """Append key/value metadata"""
    def child(self, *args, **kwargs): ...
    def finish(self, *args, **kwargs): ...
    def set_child(self, *args, **kwargs): ...
    def set_dictionary(self, *args, **kwargs): ...
    def set_dictionary_ordered(self, *args, **kwargs): ...
    def set_flags(self, *args, **kwargs): ...
    def set_format(self, *args, **kwargs): ...
    def set_name(self, *args, **kwargs): ...
    def set_nullable(self, *args, **kwargs): ...
    def set_type(self, *args, **kwargs): ...
    def set_type_date_time(self, *args, **kwargs): ...
    def set_type_decimal(self, *args, **kwargs): ...
    def set_type_fixed_size(self, *args, **kwargs): ...
    def validate(self, *args, **kwargs): ...
    def __reduce__(self): ...

class CSchemaView:
    buffer_format: Incomplete
    decimal_bitwidth: Incomplete
    decimal_precision: Incomplete
    decimal_scale: Incomplete
    dictionary_ordered: Incomplete
    extension_metadata: Incomplete
    extension_name: Incomplete
    fixed_size: Incomplete
    layout: Incomplete
    map_keys_sorted: Incomplete
    nullable: Incomplete
    storage_buffer_format: Incomplete
    storage_type: Incomplete
    storage_type_id: Incomplete
    time_unit: Incomplete
    time_unit_id: Incomplete
    timezone: Incomplete
    type: Incomplete
    type_id: Incomplete
    union_type_ids: Incomplete
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    def __reduce__(self): ...

class SchemaMetadata:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    @classmethod
    def __init__(cls, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    @staticmethod
    def empty(*args, **kwargs):
        """Create an empty SchemaMetadata with no keys or values"""
    def items(self, *args, **kwargs):
        """Iterate over key/value pairs

        The result may contain duplicate keys if they exist in the metadata."""
    def keys(self, *args, **kwargs):
        """List meadata keys

        The result may contain duplicate keys if they exist in the metadata.
        """
    def values(self, *args, **kwargs):
        """List metadata values"""
    def __contains__(self, other) -> bool:
        """Return bool(key in self)."""
    def __getitem__(self, index):
        """Get the value associated with a unique key

        Retrieves the unique value associated with k. Raises KeyError if
        k does not point to exactly one value in the metadata.
        """
    def __iter__(self):
        """Implement iter(self)."""
    def __len__(self) -> int:
        """Return len(self)."""
    def __reduce__(self): ...



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import pytest
from nanoarrow.c_array_stream import CArrayStream

import nanoarrow as na


def test_c_array_stream_from_c_array_stream():
    # Wrapping an existing stream is a no-op
    array_stream = CArrayStream.from_c_arrays([], na.c_schema(na.int32()))
    stream_from_stream = na.c_array_stream(array_stream)
    assert stream_from_stream is array_stream

    # With requested_schema should go through capsule
    array_stream = CArrayStream.from_c_arrays([], na.c_schema(na.int32()))
    with pytest.raises(NotImplementedError):
        na.c_array_stream(array_stream, na.int64())


def test_c_array_stream_from_capsule_protocol():
    # Use wrapper object to ensure this is the path taken in the constructor
    class CArrayStreamWrapper:
        def __init__(self, obj):
            self.obj = obj

        def __arrow_c_stream__(self, *args, **kwargs):
            return self.obj.__arrow_c_stream__(*args, **kwargs)

    array_stream = CArrayStream.from_c_arrays([], na.c_schema(na.int32()))
    array_stream_wrapper = CArrayStreamWrapper(array_stream)
    from_protocol = na.c_array_stream(array_stream_wrapper)
    assert array_stream.is_valid() is False
    assert from_protocol.get_schema().format == "i"


def test_c_array_stream_from_old_pyarrow():
    # Simulate a pyarrow RecordBatchReader with no __arrow_c_stream__
    class MockLegacyPyarrowRecordBatchReader:
        def __init__(self, obj):
            self.obj = obj

        def _export_to_c(self, *args):
            return self.obj._export_to_c(*args)

    MockLegacyPyarrowRecordBatchReader.__module__ = "pyarrow.lib"

    pa = pytest.importorskip("pyarrow")
    reader = pa.RecordBatchReader.from_batches(pa.schema([]), [])
    mock_reader = MockLegacyPyarrowRecordBatchReader(reader)

    array_stream = na.c_array_stream(mock_reader)
    assert array_stream.get_schema().format == "+s"


def test_c_array_stream_from_bare_capsule():
    array_stream = CArrayStream.from_c_arrays([], na.c_schema(na.int32()))

    # Check from bare capsule without supplying a schema
    capsule = array_stream.__arrow_c_stream__()
    from_capsule = na.c_array_stream(capsule)
    assert from_capsule.get_schema().format == "i"

    array_stream = CArrayStream.from_c_arrays([], na.c_schema(na.int32()))
    capsule = array_stream.__arrow_c_stream__()

    with pytest.raises(TypeError, match="Can't import c_array_stream"):
        na.c_array_stream(capsule, na.int32())


def test_c_array_stream_from_c_array_fallback():
    # Check that arrays are valid input
    c_array = na.c_array([1, 2, 3], na.int32())
    array_stream = na.c_array_stream(c_array)
    assert array_stream.get_schema().format == "i"
    arrays = list(array_stream)
    assert len(arrays) == 1
    assert arrays[0].buffers == c_array.buffers

    # Check fallback with schema
    array_stream = na.c_array_stream([1, 2, 3], na.int32())
    assert array_stream.get_schema().format == "i"
    arrays = list(array_stream)
    assert len(arrays) == 1


def test_c_array_stream_error():
    msg = "Can't resolve ArrayBuilder"
    with pytest.raises(TypeError, match=msg):
        na.c_array_stream(None)


def test_array_stream_from_arrays_schema():
    schema_in = na.c_schema(na.int32())

    stream = CArrayStream.from_c_arrays([], schema_in)
    assert schema_in.is_valid()
    assert list(stream) == []
    assert stream.get_schema().format == "i"

    # Check move of schema
    CArrayStream.from_c_arrays([], schema_in, move=True)
    assert schema_in.is_valid() is False
    assert stream.get_schema().format == "i"


def test_array_stream_from_arrays():
    schema_in = na.c_schema(na.int32())
    array_in = na.c_array([1, 2, 3], na.int32())
    array_in_buffers = array_in.buffers

    stream = CArrayStream.from_c_arrays([array_in], schema_in)
    assert array_in.is_valid()
    arrays = list(stream)
    assert len(arrays) == 1
    assert arrays[0].buffers == array_in_buffers

    # Check move of array
    stream = CArrayStream.from_c_arrays([array_in], schema_in, move=True)
    assert array_in.is_valid() is False
    arrays = list(stream)
    assert len(arrays) == 1
    assert arrays[0].buffers == array_in_buffers


def test_array_stream_from_arrays_validate():
    schema_in = na.c_schema(na.null())
    array_in = na.c_array([1, 2, 3], na.int32())

    # Check that we can skip validation and proceed without error
    stream = CArrayStream.from_c_arrays([array_in], schema_in, validate=False)
    arrays = list(stream)
    assert len(arrays) == 1
    assert arrays[0].n_buffers == 2

    # ...but that validation does happen by default
    with pytest.raises(ValueError, match="Expected schema"):
        CArrayStream.from_c_arrays([array_in], schema_in)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import pytest
from nanoarrow._buffer import CBuffer
from nanoarrow._utils import obj_is_capsule

import nanoarrow as na

np = pytest.importorskip("numpy")


def check_dlpack_export(view, expected_arr):
    # Check device spec
    assert view.__dlpack_device__() == (1, 0)

    # Check capsule export
    capsule = view.__dlpack__()
    assert obj_is_capsule(capsule, "dltensor") is True

    # Check roundtrip through numpy
    result = np.from_dlpack(view)
    np.testing.assert_array_equal(result, expected_arr, strict=True)

    # Check roundtrip through CBuffer
    buffer = CBuffer.from_dlpack(view)
    np.testing.assert_array_equal(np.array(buffer), expected_arr, strict=True)


@pytest.mark.parametrize(
    ("value_type", "np_type"),
    [
        (na.uint8(), np.uint8),
        (na.uint16(), np.uint16),
        (na.uint32(), np.uint32),
        (na.uint64(), np.uint64),
        (na.int8(), np.int8),
        (na.int16(), np.int16),
        (na.int32(), np.int32),
        (na.int64(), np.int64),
        (na.interval_months(), np.int32),
        (na.float16(), np.float16),
        (na.float32(), np.float32),
        (na.float64(), np.float64),
    ],
)
def test_dlpack(value_type, np_type):
    if np.__version__ < "1.24.0":
        pytest.skip(
            "No dlpack support in numpy versions older than 1.22.0, "
            "strict keyword in assert_array_equal added in numpy version "
            "1.24.0"
        )

    expected = np.array([1, 2, 3], dtype=np_type)
    # Use the value buffer of the nanoarrow CArray
    view = na.c_array([1, 2, 3], value_type).view().buffer(1)
    check_dlpack_export(view, expected)


def test_dlpack_not_supported():
    # DLPack doesn't support bit-packed boolean values
    view = na.c_array([True, False, True], na.bool_()).view().buffer(1)

    with pytest.raises(
        ValueError, match="Bit-packed boolean data type not supported by DLPack."
    ):
        view.__dlpack__()

    with pytest.raises(
        ValueError, match="Bit-packed boolean data type not supported by DLPack."
    ):
        view.__dlpack_device__()


def test_dlpack_cuda(cuda_device):
    cp = pytest.importorskip("cupy")
    if not cuda_device:
        pytest.skip("CUDA device not available")

    gpu_array = cp.array([1, 2, 3])
    gpu_buffer = na.c_buffer(gpu_array)
    assert gpu_buffer.device == cuda_device

    gpu_array_roundtrip = cp.from_dlpack(gpu_buffer.view())
    cp.testing.assert_array_equal(gpu_array_roundtrip, gpu_array)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# The functions here are imported from Cython. They're defined here
# instead of there to make it easier to iterate (no need to rebuild
# after editing when working with an editable installation)


def make_class_label(obj, module=None):
    if module is None:
        module = obj.__class__.__module__
    return f"{module}.{obj.__class__.__name__}"


def c_schema_to_string(obj, max_char_width=80):
    c_schema_string = ""
    if max_char_width == 0:
        c_schema_string = obj._to_string(recursive=True, max_chars=max_char_width)
    else:
        max_char_width = max(max_char_width, 10)
        c_schema_string = obj._to_string(recursive=True, max_chars=max_char_width + 1)
        if len(c_schema_string) > max_char_width:
            c_schema_string = c_schema_string[: (max_char_width - 3)] + "..."
    return c_schema_string


def metadata_repr(obj, indent=0, max_char_width=80):
    indent_str = " " * indent
    lines = []
    for key, value in obj.items():
        line = f"{indent_str}- {repr(key)}: {repr(value)}"
        lines.append(line[:max_char_width])

    return "\n".join(lines)


def schema_repr(schema, indent=0):
    indent_str = " " * indent
    class_label = make_class_label(schema, module="nanoarrow.c_schema")
    if schema._addr() == 0:
        return f"<{class_label} <NULL>>"
    elif not schema.is_valid():
        return f"<{class_label} <released>>"

    lines = [f"<{class_label} {schema._to_string()}>"]

    for attr in ("format", "name", "flags"):
        attr_repr = repr(getattr(schema, attr))
        lines.append(f"{indent_str}- {attr}: {attr_repr}")

    metadata = schema.metadata
    if schema.metadata is None:
        lines.append(f"{indent_str}- metadata: NULL")
    else:
        lines.append(f"{indent_str}- metadata:")
        lines.append(metadata_repr(metadata, indent + 2))

    if schema.dictionary:
        dictionary_repr = schema_repr(schema.dictionary, indent=indent + 2)
        lines.append(f"{indent_str}- dictionary: {dictionary_repr}")
    else:
        lines.append(f"{indent_str}- dictionary: NULL")

    lines.append(f"{indent_str}- children[{schema.n_children}]:")
    for child in schema.children:
        child_repr = schema_repr(child, indent=indent + 4)
        lines.append(f"{indent_str}  {repr(child.name)}: {child_repr}")

    return "\n".join(lines)


def array_repr(array, indent=0, max_char_width=80):
    if max_char_width < 20:
        max_char_width = 20

    indent_str = " " * indent
    class_label = make_class_label(array, module="nanoarrow.c_array")
    if array._addr() == 0:
        return f"<{class_label} <NULL>>"
    elif not array.is_valid():
        return f"<{class_label} <released>>"

    schema_string = array.schema._to_string(
        max_chars=max_char_width - indent - 23, recursive=True
    )
    lines = [f"<{class_label} {schema_string}>"]
    for attr in ("length", "offset", "null_count", "buffers"):
        attr_repr = repr(getattr(array, attr))
        lines.append(f"{indent_str}- {attr}: {attr_repr}")

    if array.dictionary:
        dictionary_repr = array_repr(array.dictionary, indent=indent + 2)
        lines.append(f"{indent_str}- dictionary: {dictionary_repr}")
    else:
        lines.append(f"{indent_str}- dictionary: NULL")

    lines.append(f"{indent_str}- children[{array.n_children}]:")
    for child in array.children:
        child_repr = array_repr(child, indent=indent + 4)
        lines.append(f"{indent_str}  {repr(child.schema.name)}: {child_repr}")

    return "\n".join(lines)


def schema_view_repr(schema_view):
    class_label = make_class_label(schema_view, module="nanoarrow.c_schema")

    lines = [
        f"<{class_label}>",
        f"- type: {repr(schema_view.type)}",
        f"- storage_type: {repr(schema_view.storage_type)}",
    ]

    for attr_name in sorted(dir(schema_view)):
        if attr_name.startswith("_") or attr_name in ("type", "storage_type"):
            continue

        attr_value = getattr(schema_view, attr_name)
        if attr_value is None:
            continue

        lines.append(f"- {attr_name}: {repr(attr_value)}")

    return "\n".join(lines)


def array_view_repr(array_view, max_char_width=80, indent=0):
    indent_str = " " * indent
    class_label = make_class_label(array_view, module="nanoarrow.c_array")

    lines = [f"<{class_label}>"]

    for attr in ("storage_type", "length", "offset", "null_count"):
        attr_repr = repr(getattr(array_view, attr))
        lines.append(f"{indent_str}- {attr}: {attr_repr}")

    lines.append(f"{indent_str}- buffers[{array_view.n_buffers}]:")
    for i, buffer in enumerate(array_view.buffers):
        buffer_type = array_view.buffer_type(i)
        lines.append(
            f"{indent_str}  - {buffer_type} "
            f"<{buffer_view_repr(buffer, max_char_width - indent - 4 - len(buffer))}>"
        )

    if array_view.dictionary:
        dictionary_repr = array_view_repr(
            array_view.dictionary, max_char_width=max_char_width, indent=indent + 2
        )
        lines.append(f"{indent_str}- dictionary: {dictionary_repr}")
    else:
        lines.append(f"{indent_str}- dictionary: NULL")

    lines.append(f"{indent_str}- children[{array_view.n_children}]:")
    for child in array_view.children:
        child_repr = array_view_repr(
            child, max_char_width=max_char_width, indent=indent + 4
        )
        lines.append(f"{indent_str}  - {child_repr}")

    return "\n".join(lines)


def buffer_view_repr(buffer_view, max_char_width=80):
    if max_char_width < 20:
        max_char_width = 20

    prefix = f"{buffer_view.data_type}"
    prefix += f"[{buffer_view.size_bytes} b]"

    if buffer_view.device.device_type_id == 1:
        return (
            prefix
            + " "
            + buffer_view_preview_cpu(buffer_view, max_char_width - len(prefix) - 2)
        )
    else:
        dev_info = f"<{buffer_view.device.device_type}/{buffer_view.device.device_id}>"
        return prefix + dev_info


def buffer_view_preview_cpu(buffer_view, max_char_width):
    if buffer_view.element_size_bits == 0:
        preview_elements = max_char_width - 3
        joined = repr(bytes(memoryview(buffer_view)[:preview_elements]))
    elif buffer_view.element_size_bits == 1:
        max_elements = max_char_width // 8
        if max_elements > len(buffer_view):
            preview_elements = len(buffer_view)
        else:
            preview_elements = max_elements

        joined = "".join(
            "".join(reversed(format(buffer_view[i], "08b")))
            for i in range(preview_elements)
        )
    else:
        max_elements = max_char_width // 3
        if max_elements > len(buffer_view):
            preview_elements = len(buffer_view)
        else:
            preview_elements = max_elements

        joined = " ".join(repr(buffer_view[i]) for i in range(preview_elements))

    if len(joined) > max_char_width or preview_elements < len(buffer_view):
        return joined[: (max_char_width - 3)] + "..."
    else:
        return joined


def array_stream_repr(array_stream, max_char_width=80):
    class_label = make_class_label(array_stream, module="nanoarrow.c_array_stream")

    if array_stream._addr() == 0:
        return f"<{class_label} <NULL>>"
    elif not array_stream.is_valid():
        return f"<{class_label} <released>>"

    lines = [f"<{class_label}>"]
    try:
        schema = array_stream.get_schema()
        schema_string = schema._to_string(max_chars=max_char_width - 16, recursive=True)
        lines.append(f"- get_schema(): {schema_string}")
    except Exception as e:
        lines.append(f"- get_schema(): <error calling get_schema(): {e}>")

    return "\n".join(lines)


def device_array_repr(device_array):
    class_label = make_class_label(device_array, module="nanoarrow.device")

    title_line = f"<{class_label}>"
    device_type = (
        f"- device_type: {device_array.device_type.name} "
        f"<{device_array.device_type_id}>"
    )
    device_id = f"- device_id: {device_array.device_id}"
    array = f"- array: {array_repr(device_array.array, indent=2)}"
    return "\n".join((title_line, device_type, device_id, array))


def device_repr(device):
    class_label = make_class_label(device, module="nanoarrow.device")

    title_line = f"<{class_label}>"
    device_type = f"- device_type: {device.device_type.name} <{device.device_type_id}>"
    device_id = f"- device_id: {device.device_id}"
    return "\n".join([title_line, device_type, device_id])


def array_inspect(array, indent=0, max_char_width=80):
    array_view = array.view()

    if max_char_width < 20:
        max_char_width = 20

    indent_str = " " * indent
    class_label = "ArrowArray"
    if array._addr() == 0:
        return f"<{class_label} <NULL>>"
    elif not array.is_valid():
        return f"<{class_label} <released>>"

    schema_string = array.schema._to_string(
        max_chars=max_char_width - indent - 23, recursive=True
    )
    lines = [f"<{class_label} {schema_string}>"]
    for attr in ("length", "offset", "null_count"):
        attr_repr = repr(getattr(array, attr))
        lines.append(f"{indent_str}- {attr}: {attr_repr}")

    lines.append(f"{indent_str}- buffers[{array_view.n_buffers}]:")
    for i, buffer in enumerate(array_view.buffers):
        buffer_type = array_view.buffer_type(i)
        lines.append(
            f"{indent_str}  - {buffer_type} "
            f"<{buffer_view_repr(buffer, max_char_width - indent - 4 - len(buffer))}>"
        )

    if array.dictionary:
        dictionary_repr = array_inspect(array.dictionary, indent=indent + 2)
        lines.append(f"{indent_str}- dictionary: {dictionary_repr}")
    else:
        lines.append(f"{indent_str}- dictionary: NULL")

    lines.append(f"{indent_str}- children[{array.n_children}]:")
    for child in array.children:
        child_repr = array_inspect(child, indent=indent + 4)
        lines.append(f"{indent_str}  {repr(child.schema.name)}: {child_repr}")

    return "\n".join(lines)
